#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------
"""
Title: UVMPair classes

This section defines container classes for handling value pairs.

"""

from uvm.base.sv import sv
from uvm.base.uvm_object import UVMObject
from uvm.macros import uvm_object_utils, uvm_error



class UVMClassPair(UVMObject):
    """
    Class: UVMClassPair

    Container holding handles to two objects whose types are specified by the
    type parameters, T1 and T2.
    """


    type_name = "UVMClassPair"

    #// Variable: T1 first
    #//
    #// The handle to the first object in the pair
    #
    #  T1 first
    #
    #// Variable: T2 second
    #//
    #// The handle to the second object in the pair
    #
    #  T2 second
    #

    #  // Function: new
    #  //
    #  // Creates an instance that holds a handle to two objects.
    #  // The optional name argument gives a name to the new pair object.
    def __init__(self, name="", f=None, s=None):
        super().__init__(name)
        self.first = f
        self.second = s

        #if f is None:
        #    self.first = new
        #else
        #  first = f
        #if (s is None)
        #    second = new
        #else
        #    second = s


    #  def get_type_name(self):
    #    return type_name
    #  endfunction

    def convert2string(self):
        s = sv.sformatf("pair : %s, %s",
             self.first.convert2string(), self.second.convert2string())
        return s


    #  def do_compare(self,uvm_object rhs, uvm_comparer comparer):
    #    this_type rhs_
    #    if(!sv.cast(rhs_,rhs)):
    #      uvm_error("WRONG_TYPE", {"do_compare: rhs argument is not of type '",get_type_name(),"'"})
    #      return 0
    #    end
    #    return first.compare(rhs_.first)  and  second.compare(rhs_.second)
    #  endfunction

    #  def do_copy(self,uvm_object rhs):
    #    this_type rhs_
    #    if(!sv.cast(rhs_,rhs))
    #      uvm_fatal("WRONG_TYPE", {"do_copy: rhs argument is not of type '",get_type_name(),"'"})
    #    first.copy(rhs_.first)
    #    second.copy(rhs_.second)
    #  endfunction


uvm_object_utils(UVMClassPair)


#//-----------------------------------------------------------------------------
#// CLASS: UVMBuiltInPair #(T1,T2)
#//
#// Container holding two variables of built-in types (int, string, etc.). The
#// types are specified by the type parameters, T1 and T2.
#//-----------------------------------------------------------------------------

class UVMBuiltInPair(UVMObject):


    type_name = "UVMBuiltInPair"

    #// Variable: T1 first
    #//
    #// The first value in the pair
    #
    #  T1 first

    #// Variable: T2 second
    #//
    #// The second value in the pair
    #  T2 second


    #  // Function: new
    #  //
    #  // Creates an instance that holds two built-in type values.
    #  // The optional name argument gives a name to the new pair object.
    #
    def __init__(self, name=""):
        super().__init__(name)
        self.first = 0
        self.seconds = 0


    def get_type_name(self):
        return UVMBuiltInPair.type_name


    def convert2string(self):
        return "built-in pair : {}, {}".format(self.first, self.second)


    def do_compare(self, rhs, comparer):
        rhs_ = []
        if not sv.cast(rhs_,rhs, UVMBuiltInPair):
            uvm_error("WRONG_TYPE", {"do_compare: rhs argument is not of type '",
                self.get_type_name(),"'"})
            return 0
        rhs_ = rhs[0]
        return self.first == rhs_.first and self.second == rhs_.second


    #  def do_copy(self,uvm_object rhs):
    #    this_type rhs_
    #    if(!sv.cast(rhs_,rhs))
    #      uvm_fatal("WRONG_TYPE", {"do_copy: rhs argument is not of type '",get_type_name(),"'"})
    #    first = rhs_.first
    #    second = rhs_.second

uvm_object_utils(UVMBuiltInPair)



#
#-----------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2013 NVIDIA Corporation
#   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-----------------------------------------------------------------------------

from typing import Dict, Any, List

from .sv import sv, sv_obj
from .uvm_misc import UVMStatusContainer
from .uvm_object_globals import (UVM_PRINT, UVM_NONE, UVM_COPY, UVM_COMPARE,
        UVM_RECORD, UVM_SETINT, UVM_SETOBJ, UVM_SETSTR, UVM_PACK, UVM_UNPACK)
from .uvm_globals import uvm_report_error, uvm_report_warning, uvm_report_info
from typing import Tuple


class UVMObject(sv_obj):
    """
    The `UVMObject` class is the base class for all UVM data and hierarchical classes.
    Its primary role is to define a set of methods for such common operations as `create`,
    `copy`, `compare`, `print`, and `record`.
    Classes deriving from `UVMObject` must implement methods such as
    `create` and `get_type_name`.

    :ivar str name: Name of the object
    :ivar int inst_id: Unique instance ID for this object

    Group: Seeding

    :cvar bool use_uvm_seeding: This bit enables or disables the UVM seeding
    mechanism. It globally affects the operation of the `reseed` method.

    When enabled, UVM-based objects are seeded based on their type and full
    hierarchical name rather than allocation order. This improves random
    stability for objects whose instance names are unique across each type.
    The `UVMComponent` class is an example of a type that has a unique
    instance name.
    """

    # Should be set by uvm_*_utils macro
    type_id = None  # type: Any

    depth = 0
    m_inst_count = 0

    m_inst_count = 0
    use_uvm_seeding = True
    uvm_global_copy_map = {}  # type: Dict['UVMObject', 'UVMObject']
    _m_uvm_status_container = UVMStatusContainer()

    def __init__(self, name: str):
        """ Creates a new uvm_object with the given instance `name`. If `name` is not
        supplied, the object is unnamed.
        """
        sv_obj.__init__(self)
        self.name = name
        self.inst_id = UVMObject.m_inst_count
        UVMObject.m_inst_count += 1
        self.leaf_name = name


    def reseed(self) -> None:
        """
        Calls `srandom` on the object to reseed the object using the UVM seeding
        mechanism, which sets the seed based on type name and instance name instead
        of based on instance position in a thread.

        If the `use_uvm_seeding` static variable is set to 0, then reseed() does
        not perform any function.
        """
        if (UVMObject.use_uvm_seeding):
            pass


    """
        Group: Identification
    """

    def set_name(self, name: str):
        """
        Sets the instance name of this object, overwriting any previously
        given name.

        Args:
            name (str): Name for the object.
        """
        self.leaf_name = name

    def get_name(self) -> str:
        """
        Returns the name of the object, as provided by the `name` argument in the
        `new` constructor or `set_name` method.

        Returns:
            str: Name of the object.
        """
        return self.leaf_name

    def get_full_name(self) -> str:
        """
        Objects possessing hierarchy, such as <uvm_components>, override the default
        implementation. Other objects might be associated with component hierarchy
        but are not themselves components. For example, <uvm_sequence #(REQ,RSP)>
        classes are typically associated with a <uvm_sequencer #(REQ,RSP)>. In this
        case, it is useful to override get_full_name to return the sequencer's
        full name concatenated with the sequence's name. This provides the sequence
        a full context, which is useful when debugging.

        Returns:
            str: The full hierarchical name of this object. The default
            implementation is the same as <get_name>, as uvm_objects do not inherently
            possess hierarchy.
        """
        return self.get_name()

    def get_inst_id(self) -> int:
        """
        Returns:
            int: The object's unique, numeric instance identifier.
        """
        return self.inst_id


    @classmethod
    def get_inst_count(self) -> int:
        """
        Returns:
            int: The current value of the instance counter, which represents the
            total number of uvm_object-based objects that have been allocated in
            simulation. The instance counter is used to form a unique numeric instance
            identifier.
        """
        return UVMObject.m_inst_count

    def get_type(self) -> None:
        """
        Returns the type-proxy (wrapper) for this object. The `UVMFactory`'s
        type-based override and creation methods take arguments of
        `uvm_object_wrapper`. This method, if implemented, can be used as convenient
        means of supplying those arguments.

        The default implementation of this method produces an error and returns
        `None`. To enable use of this method, a user's subtype must implement a
        version that returns the subtype's wrapper.

        For example:

        .. code-block:: python

          class cmd(UVMObject):
            type_id = None

            @classmethod
            def get_type(cls):
              return cls.type_id.get()

        Then, to use:

        .. code-block:: python

            factory.set_type_override(cmd.get_type(), subcmd.get_type())

        This function is implemented by the uvm_*_utils functions, if employed.

        Returns:
        """
        uvm_report_error("NOTYPID", "get_type not implemented in derived class: "
                + str(self), UVM_NONE)
        return None

    def get_object_type(self) -> Any:
        """
        Function: get_object_type

        Returns the type-proxy (wrapper) for this object. The `uvm_factory`'s
        type-based override and creation methods take arguments of
        `uvm_object_wrapper`. This method, if implemented, can be used as convenient
        means of supplying those arguments. This method is the same as the static
        `get_type` method, but uses an already allocated object to determine
        the type-proxy to access (instead of using the static object).

        The default implementation of this method does a factory lookup of the
        proxy using the return value from `get_type_name`. If the type returned
        by `get_type_name` is not registered with the factory, then a `None`
        handle is returned.

        For example:

        .. code-block:: python

          class cmd (UVMObject):
            type_id = UVMObjectRegistry()
            @classmethod
            def type_id get_type(cls):
              return type_id.get()
            def get_object_type(self):
              return cmd.type_id.get()


        This function is implemented by the `uvm_*_utils macros, if employed.

        Returns:
        """
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        factory = cs.get_factory()
        if self.get_type_name() == "<unknown>":
            return None
        return factory.find_wrapper_by_name(self.get_type_name())

    def get_type_name(self) -> str:
        """
        This function returns the type name of the object, which is typically the
        type identifier enclosed in quotes. It is used for various debugging
        functions in the library, and it is used by the factory for creating
        objects.

        This function must be defined in every derived class.

        A typical implementation is as follows:

        .. code-block:: python
          class mytype (UVMObject):
            ...
            type_name = "mytype"

            def get_type_name(self):
              return my_type.type_name


          We define the `type_name` static variable to enable access to the type name
          without need of an object of the class, i.e., to enable access via the
          scope operator, ~mytype::type_name~.
        Returns:
            str: Type name of the object.
        """
        return "<unknown>"

    def create(self, name="") -> 'UVMObject':
        """
        Group: Creation

        The `create` method allocates a new object of the same type as this object
        and returns it via a base uvm_object handle. Every class deriving from
        uvm_object, directly or indirectly, must implement the create method.

        A typical implementation is as follows:

        .. code-block:: python

          class mytype (UVMObject):
            ...
            def create(self, name=""):
              mytype t = mytype(name)
              return t

        Args:
            name (str): Name of the created object.
        Returns:
            obj: New object.
        """
        return UVMObject(name)

    def clone(self) -> 'UVMObject':
        """
        The `clone` method creates and returns an exact copy of this object.

        The default implementation calls `create` followed by `copy`. As clone is
        virtual, derived classes may override this implementation if desired.

        Returns:
            UVMObject: Clone of the object.
        """
        tmp = self.create(self.get_name())
        if tmp is None:
            uvm_report_warning("CRFLD", sv.sformatf(
                "The create method failed for %s,  object cannot be cloned",
                self.get_name()), UVM_NONE)
        else:
            tmp.copy(self)
        return tmp

    def print_obj(self, printer=None) -> None:
        """
        Group: Printing

        Function: print

        The `print` method deep-prints this object's properties in a format and
        manner governed by the given `printer` argument; if the `printer` argument
        is not provided, the global `uvm_default_printer` is used. See
        `uvm_printer` for more information on printer output formatting. See also
        `uvm_line_printer`, `uvm_tree_printer`, and `uvm_table_printer` for details
        on the pre-defined printer "policies," or formatters, provided by the UVM.

        The `print` method is not virtual and must not be overloaded. To include
        custom information in the `print` and `sprint` operations, derived classes
        must override the `do_print` method and use the provided printer policy
        class to format the output.

        Args:
            printer (UVMPrinter): Printer that is used in printing.
        """
        if printer is None:
            from .uvm_global_vars import uvm_default_printer
            printer = uvm_default_printer
        if printer is None:
            uvm_report_error("NonePRINTER", "uvm_default_printer is None")
        sv.fwrite(printer.knobs.mcd, self.sprint(printer))

    def sprint(self, printer=None) -> str:
        """
        The `sprint` method works just like the `print` method, except the output
        is returned in a string rather than displayed.

        The `sprint` method is not virtual and must not be overloaded. To include
        additional fields in the `print` and `sprint` operation, derived classes
        must override the `do_print` method and use the provided printer policy
        class to format the output. The printer policy will manage all string
        concatenations and provide the string to `sprint` to return to the caller.

        Args:
            printer (UVMPrinter): Printer that is used in printing.
        Returns:
            str: String representation of the object.
        """
        if printer is None:
            from .uvm_global_vars import uvm_default_printer
            printer = uvm_default_printer
        if not printer.istop():
            UVMObject._m_uvm_status_container.printer = printer
            self._m_uvm_field_automation(None, UVM_PRINT, "")
            self.do_print(printer)
            return ""
        self._m_uvm_status_container = UVMObject._m_uvm_status_container
        printer.print_object(self.get_name(), self)
        if printer.m_string != "":
            return printer.m_string
        return printer.emit()

    def do_print(self, printer) -> None:
        """
        The `do_print` method is the user-definable hook called by `print` and
        `sprint` that allows users to customize what gets printed or sprinted
        beyond the field information provided by the `uvm_field_* macros,
        <Utility and Field Macros for Components and Objects>.

        The `printer` argument is the policy object that governs the format and
        content of the output. To ensure correct `print` and `sprint` operation,
        and to ensure a consistent output format, the `printer` must be used
        by all `do_print` implementations. That is, instead of using ~$display~ or
        string concatenations directly, a `do_print` implementation must call
        through the ~printer's~ API to add information to be printed or sprinted.

        An example implementation of `do_print` is as follows::

          class mytype (UVMObject):
            data_obj data
            int f1
            virtual function void do_print (uvm_printer printer)
              super.do_print(printer)
              printer.print_field_int("f1", f1, $bits(f1), UVM_DEC)
              printer.print_object("data", data)
            endfunction

        Then, to print and sprint the object, you could write::

          t = mytype()
          t.print()
          uvm_report_info("Received",t.sprint())

        See `UVMPrinter` for information about the printer API.

        Args:
            printer (UVMPrinter): Printer that is used in printing.
        """
        return

    def convert2string(self) -> str:
        """
        This virtual function is a user-definable hook, called directly by the
        user, that allows users to provide object information in the form of
        a string. Unlike `sprint`, there is no requirement to use a `uvm_printer`
        policy object. As such, the format and content of the output is fully
        customizable, which may be suitable for applications not requiring the
        consistent formatting offered by the `print`/`sprint`/`do_print`
        API.

        Fields declared in <Utility Macros> macros (`uvm_field_*), if used, will
        not automatically appear in calls to convert2string.

        An example implementation of convert2string follows.

        .. code-block:: python

          class Base(UVMObject):
            field = "foo"
            def convert2string(self):
              return "base_field=" + self.field

          class Obj2(UVMObject):
            field = "bar"
            def convert2string()
              convert2string = "child_field=" + self.field

          class Obj(Base):
            addr = 0x123
            data = 0x456
            write = 1
            child = Obj2()
            def convert2string(self):
               convert2string = super().convert2string() +
                 sv.sformatf(" write=%0d addr=%8h data=%8h ",write,addr,data) +
                 child.convert2string()

         Then, to display an object, you could write:

         .. code-block:: python

           o = Obj()
           uvm_report_info("BusMaster", "Sending:\n " + o.convert2string())

        The output will look similar to::

          UVM_INFO @ 0: reporter [BusMaster] Sending:
            base_field=foo write=1 addr=00000123 data=00000456 child_field=bar

        Returns:
            str: Object converted into string.
        """
        return ""

    def _m_uvm_field_automation(self, tmp_data__, what__, str__) -> None:
        pass

    def record(self, recorder=None) -> None:
        """
        Group: Recording

        The `record` method deep-records this object's properties according to an
        optional `recorder` policy. The method is not virtual and must not be
        overloaded. To include additional fields in the record operation, derived
        classes should override the `do_record` method.

        The optional `recorder` argument specifies the recording policy, which
        governs how recording takes place. See
        `uvm_recorder` for information.

        A simulator's recording mechanism is vendor-specific. By providing access
        via a common interface, the uvm_recorder policy provides vendor-independent
        access to a simulator's recording capabilities.

        Args:
            recorder (UVMRecorder):
        """
        if recorder is None:
            return
        UVMObject._m_uvm_status_container.recorder = recorder
        recorder.recording_depth += 1
        self._m_uvm_field_automation(None, UVM_RECORD, "")
        self.do_record(recorder)
        recorder.recording_depth -= 1


    def do_record(self, recorder) -> None:
        """
        The `do_record` method is the user-definable hook called by the `record`
        method. A derived class should override this method to include its fields
        in a record operation.

        The `recorder` argument is policy object for recording this object. A
        do_record implementation should call the appropriate recorder methods for
        each of its fields. Vendor-specific recording implementations are
        encapsulated in the `recorder` policy, thereby insulating user-code from
        vendor-specific behavior. See `uvm_recorder` for more information.

        A typical implementation is as follows:

        .. code-block:: python

          class mytype (UVMObject):
            data_obj data
            int f1
            def do_record (self, recorder):
              recorder.record_field("f1", f1, sv.bits(f1), UVM_DEC)
              recorder.record_object("data", data)

        Args:
            recorder (UVMRecorder): Recorder policy object.
        """
        return

    def copy(self, rhs: 'UVMObject'):
        """
        The copy makes this object a copy of the specified object.

        The `copy` method is not virtual and should not be overloaded in derived
        classes. To copy the fields of a derived class, that class should override
        the `do_copy` method.

        Args:
            rhs (UVMObject): An object to be copied.
        """
        # For cycle checking
        UVMObject.depth = 0
        if (rhs is not None) and rhs in UVMObject.uvm_global_copy_map:
            return

        if rhs is None:
            uvm_report_warning("NoneCP",
                "A None object was supplied to copy; copy is ignored", UVM_NONE)
            return

        UVMObject.uvm_global_copy_map[rhs] = self
        UVMObject.depth += 1
        self._m_uvm_field_automation(rhs, UVM_COPY, "")
        self.do_copy(rhs)

        UVMObject.depth -= 1
        if UVMObject.depth == 0:
            UVMObject.uvm_global_copy_map = {}

    def do_copy(self, rhs) -> None:
        """
        The `do_copy` method is the user-definable hook called by the `copy` method.
        A derived class should override this method to include its fields in a `copy`
        operation.

        A typical implementation is as follows:

        .. code-block:: python

          class mytype (UVMObject):
            ...
            field_1 = 0
            def do_copy(self, rhs):
              super.do_copy(rhs)
              # Optionanl type checking
              field_1 = rhs.field_1

        The implementation must call `super().do_copy`, and can optionally do
        type checking before copying.

        Args:
            rhs (UVMObject): Object to be copied.
        """
        return

    def compare(self, rhs, comparer=None) -> bool:
        """
        Deep compares members of this data object with those of the object provided
        in the `rhs` (right-hand side) argument, returning 1 on a match, 0 otherwise.

        The `compare` method is not virtual and should not be overloaded in derived
        classes. To compare the fields of a derived class, that class should
        override the `do_compare` method.

        The optional `comparer` argument specifies the comparison policy. It allows
        you to control some aspects of the comparison operation. It also stores the
        results of the comparison, such as field-by-field miscompare information
        and the total number of miscompares. If a compare policy is not provided,
        then the global `uvm_default_comparer` policy is used. See `uvm_comparer`
        for more information.

        Args:
            rhs (UVMObject): Object to be compared against.
            comparer (UVMComparer): Comparer policy object.
        Returns:
            bool: True if objects match, False otherwise.
        """
        # t = 0
        dc = 0
        #static int style
        # style = 0
        done = 0
        cls = UVMObject
        if comparer is not None:
            cls._m_uvm_status_container.comparer = comparer
        else:
            from .uvm_global_vars import uvm_default_comparer
            cls._m_uvm_status_container.comparer = uvm_default_comparer
        comparer = cls._m_uvm_status_container.comparer

        if(not cls._m_uvm_status_container.scope.depth()):
            comparer.compare_map.delete()
            comparer.result = 0
            comparer.miscompares = ""
            comparer.scope = cls._m_uvm_status_container.scope
            if self.get_name() == "":
                cls._m_uvm_status_container.scope.down("<object>")
            else:
                cls._m_uvm_status_container.scope.down(self.get_name())

        if(not done and (rhs is None)):
            if(cls._m_uvm_status_container.scope.depth()):
                comparer.print_msg_object(self, rhs)
            else:
                comparer.print_msg_object(self, rhs)
                uvm_report_info("MISCMP",
                     sv.sformatf("%0d Miscompare(s) for object %s@%0d vs. None",
                     comparer.result,
                     cls._m_uvm_status_container.scope.get(),
                      self.get_inst_id()),
                      cls._m_uvm_status_container.comparer.verbosity)
                done = 1

        if(not done and comparer.compare_map.exists(rhs)):
            if(comparer.compare_map[rhs] != self):
                comparer.print_msg_object(self, comparer.compare_map[rhs])
            done = 1  # don't do any more work after this case, but do cleanup

        if(not done and comparer.check_type and (rhs is not None) and
                (self.get_type_name() != rhs.get_type_name())):
            cls._m_uvm_status_container.stringv = ("lhs type = \"" + self.get_type_name()
                + "' : rhs type = '" + rhs.get_type_name() + "'")
            comparer.print_msg(cls._m_uvm_status_container.stringv)

        if not done:
            comparer.compare_map[rhs] = self
            self._m_uvm_field_automation(rhs, UVM_COMPARE, "")
            dc = self.do_compare(rhs, comparer)

        if cls._m_uvm_status_container.scope.depth() == 1:
            cls._m_uvm_status_container.scope.up()

        if rhs is not None:
            comparer.print_rollup(self, rhs)
        return (comparer.result == 0 and dc == 1)


    def do_compare(self, rhs, comparer) -> bool:
        """
        The `do_compare` method is the user-definable hook called by the `compare`
        method. A derived class should override this method to include its fields
        in a compare operation. It should return 1 if the comparison succeeds, 0
        otherwise.

        A typical implementation is as follows:

        .. code-block:: python

          class mytype (UVMObject):
            ...
            f1 = 0
            def do_compare(self, rhs, comparer):
              do_compare = super.do_compare(rhs,comparer)
              # Optional type checking
              do_compare &= comparer.compare_field_int("f1", f1, rhs.f1)
              return do_compare

        A derived class implementation must call `super().do_compare` to ensure its
        base class' properties, if any, are included in the comparison. If type
        matching is required instead of duck-typing, the user can also
        implemented this checking.

        The actual comparison should be implemented using the `UVMComparer` object
        rather than direct field-by-field comparison. This enables users of your
        class to customize how comparisons are performed and how much miscompare
        information is collected. See `UVMComparer` for more details.

        Args:
            rhs (UVMObject):
            comparer (UVMComparer):
        Returns:
            bool: True if objects match, False otherwise.
        """
        return True


    #  // Group: Packing

    #  // Function: pack
    #
    #  extern function int pack (ref bit bitstream[],
    #                            input uvm_packer packer=None)
    def pack(self, packer=None) -> Tuple[Any, Any]:
        packer = self.m_pack(packer)
        return packer.get_packed_size(), packer.get_bits()


    #  // Function: pack_bytes
    #
    #  extern function int pack_bytes (ref byte unsigned bytestream[],
    #                                  input uvm_packer packer=None)
    def pack_bytes(self, bytestream, packer=None) -> Any:
        packer = self.m_pack(packer)
        packed_bytes = packer.get_bytes()
        for b in packed_bytes:
            bytestream.append(b)
        return packer.get_packed_size()


    def pack_ints(self, intstream: List, packer=None) -> Any:
        """
        Function: pack_ints

        The pack methods bitwise-concatenate this object's properties into an array
        of bits, bytes, or ints. The methods are not virtual and must not be
        overloaded. To include additional fields in the pack operation, derived
        classes should override the <do_pack> method.

        The optional `packer` argument specifies the packing policy, which governs
        the packing operation. If a packer policy is not provided, the global
        <uvm_default_packer> policy is used. See <uvm_packer> for more information.

        The return value is the total number of bits packed into the given array.
        Use the array's built-in `size` method to get the number of bytes or ints
        consumed during the packing process.
        """
        #
        #  extern function int pack_ints (ref int unsigned intstream[],
        #                                 input uvm_packer packer=None)
        packer = self.m_pack(packer)
        ints = packer.get_ints()
        for i in ints:
            intstream.append(i)
        return packer.get_packed_size()


    def do_pack(self, packer) -> None:
        """
        Function: do_pack

        The `do_pack` method is the user-definable hook called by the <pack> methods.
        A derived class should override this method to include its fields in a pack
        operation.

        The `packer` argument is the policy object for packing. The policy object
        should be used to pack objects.

        A typical example of an object packing itself is as follows

        .. code-block:: python

          class mysubtype(mysupertype):
            ...
            # shortint myshort
            # obj_type myobj
            # byte myarray[]
            ...
            function void do_pack (uvm_packer packer)
              super.do_pack(packer); // pack mysupertype properties
              packer.pack_field_int(len(myarray), 32)
              for index in range(len(myarray):
                packer.pack_field_int(myarray[index], 8)
              packer.pack_field_int(myshort, sv.bits(myshort))
              packer.pack_object(myobj)
            endfunction

        The implementation must call ~super.do_pack~ so that base class properties
        are packed as well.

        If your object contains dynamic data (object, string, queue, dynamic array,
        or associative array), and you intend to unpack into an equivalent data
        structure when unpacking, you must include meta-information about the
        dynamic data when packing as follows.

         - For queues, dynamic arrays, or associative arrays, pack the number of
           elements in the array in the 32 bits immediately before packing
           individual elements, as shown above.

         - For string data types, append a zero byte after packing the string
           contents.

         - For objects, pack 4 bits immediately before packing the object. For `None`
           objects, pack 4'b0000. For non-`None` objects, pack 4'b0001.

        When the uvm_field_* macros are used,
        <Utility and Field Macros for Components and Objects>,
        the above meta information is included provided the <uvm_packer::use_metadata>
        variable is set for the packer.

        Packing order does not need to match declaration order. However, unpacking
        order must match packing order.
        """
        return


    #  // Group: Unpacking
    #
    #  // Function: unpack
    #
    #  extern function int unpack (ref   bit        bitstream[],
    #                              input uvm_packer packer=None)
    def unpack(self, bitstream, packer=None) -> Any:
        packer = self.m_unpack_pre(packer)
        packer.put_bits(bitstream)
        self.m_unpack_post(packer)
        packer.set_packed_size()
        return packer.get_packed_size()


    #  // Function: unpack_bytes
    #
    #  extern function int unpack_bytes (ref byte unsigned bytestream[],
    #                                    input uvm_packer packer=None)
    def unpack_bytes(self, bytestream, packer=None) -> Any:
        packer = self.m_unpack_pre(packer)
        packer.put_bytes(bytestream)
        self.m_unpack_post(packer)
        packer.set_packed_size()
        return packer.get_packed_size()


    def unpack_ints(self, intstream, packer=None) -> Any:
        """
        The unpack methods extract property values from an array of bits, bytes, or
        ints. The method of unpacking `must` exactly correspond to the method of
        packing. This is assured if (a) the same `packer` policy is used to pack
        and unpack, and (b) the order of unpacking is the same as the order of
        packing used to create the input array.

        The unpack methods are fixed (non-virtual) entry points that are directly
        callable by the user. To include additional fields in the <unpack>
        operation, derived classes should override the <do_unpack> method.

        The optional `packer` argument specifies the packing policy, which governs
        both the pack and unpack operation. If a packer policy is not provided,
        then the global `uvm_default_packer` policy is used. See uvm_packer for
        more information.

        The return value is the actual number of bits unpacked from the given array.

        Args:
            instream (List):
            packer (UVMPacker):

        Returns:
            int: Packed size
        """
        packer = self.m_unpack_pre(packer)
        packer.put_ints(intstream)
        self.m_unpack_post(packer)
        packer.set_packed_size()
        return packer.get_packed_size()


    #  // Function: do_unpack
    #  //
    #  // The `do_unpack` method is the user-definable hook called by the <unpack>
    #  // method. A derived class should override this method to include its fields
    #  // in an unpack operation.
    #  //
    #  // The `packer` argument is the policy object for both packing and unpacking.
    #  // It must be the same packer used to pack the object into bits. Also,
    #  // do_unpack must unpack fields in the same order in which they were packed.
    #  // See <uvm_packer> for more information.
    #  //
    #  // The following implementation corresponds to the example given in do_pack.
    #  //
    #  //|  function void do_unpack (uvm_packer packer)
    #  //|   int sz
    #  //|    super.do_unpack(packer); // unpack super's properties
    #  //|    sz = packer.unpack_field_int(myarray.size(), 32)
    #  //|    myarray.delete()
    #  //|    for(int index=0; index<sz; index++)
    #  //|      myarray[index] = packer.unpack_field_int(8)
    #  //|    myshort = packer.unpack_field_int($bits(myshort))
    #  //|    packer.unpack_object(myobj)
    #  //|  endfunction
    #  //
    #  // If your object contains dynamic data (object, string, queue, dynamic array,
    #  // or associative array), and you intend to <unpack> into an equivalent data
    #  // structure, you must have included meta-information about the dynamic data
    #  // when it was packed.
    #  //
    #  // - For queues, dynamic arrays, or associative arrays, unpack the number of
    #  //   elements in the array from the 32 bits immediately before unpacking
    #  //   individual elements, as shown above.
    #  //
    #  // - For string data types, unpack into the new string until a `None` byte is
    #  //   encountered.
    #  //
    #  // - For objects, unpack 4 bits into a byte or int variable. If the value
    #  //   is 0, the target object should be set to `None` and unpacking continues to
    #  //   the next property, if any. If the least significant bit is 1, then the
    #  //   target object should be allocated and its properties unpacked.
    def do_unpack(self, packer) -> None:
        return

    def set_int_local(self, field_name: str, value: int, recurse=True):
        """
        Group: Configuration

        Args:
            field_name (str): Variable to set
            value: Value for the variable
            recurse (bool):
        """
        UVMObject._m_uvm_status_container.cycle_check.clear()
        UVMObject._m_uvm_status_container.m_uvm_cycle_scopes.clear()

        UVMObject._m_uvm_status_container.status = False
        UVMObject._m_uvm_status_container.bitstream = value

        self._m_uvm_field_automation(None, UVM_SETINT, field_name)

        if UVMObject._m_uvm_status_container.warning and not self._m_uvm_status_container.status:
            uvm_report_error("NOMTC", sv.sformatf("did not find a match for field %s", field_name),UVM_NONE)
        UVMObject._m_uvm_status_container.cycle_check.clear()


    def set_string_local(self, field_name: str, value: str, recurse=True):
        """
        Function: set_string_local

        Args:
            field_name (str): Variable to set
            value: Value for the variable
            recurse (bool): If True, recurse into sub-objects.
        """
        UVMObject._m_uvm_status_container.cycle_check.clear()
        UVMObject._m_uvm_status_container.m_uvm_cycle_scopes.clear()

        UVMObject._m_uvm_status_container.status = False
        UVMObject._m_uvm_status_container.stringv = value
        self._m_uvm_field_automation(None, UVM_SETSTR, field_name)

        if UVMObject._m_uvm_status_container.warning and not UVMObject._m_uvm_status_container.status:
            uvm_report_error("NOMTC", sv.sformatf("did not find a match for field %s (@%0d)",
                field_name, self.get_inst_id()), UVM_NONE)

        UVMObject._m_uvm_status_container.cycle_check.clear()


    def set_object_local(self, field_name: str, value: 'UVMObject', clone=1, recurse=1):
        """
        These methods provide write access to integral, string, and
        uvm_object-based properties indexed by a `field_name` string. The object
        designer choose which, if any, properties will be accessible, and overrides
        the appropriate methods depending on the properties' types. For objects,
        the optional `clone` argument specifies whether to clone the `value`
        argument before assignment.

        The global `uvm_is_match` function is used to match the field names, so
        `field_name` may contain wildcards.

        An example implementation of all three methods is as follows.

        .. code-block:: python

          class mytype(UVMObject):

            def __init__(self, name):
              super().__init__(name)
              self.myint = 0
              self.mybyte = 0
              self.myshort = 0
              self.mystring = ""
              self.myobj = None

            # provide access to integral properties
            def set_int_local(self, field_name, value):
              if (uvm_is_match (field_name, "myint")):
                self.myint = value
              elif (uvm_is_match (field_name, "mybyte")):
                selef.mybyte = value

            # provide access to string properties
            def set_string_local(self, field_name, value):
              if (uvm_is_match (field_name, "mystring")):
                self.mystring = value

            # provide access to sub-objects
            def set_object_local(self, field_name, value,clone=1):
              if (uvm_is_match (field_name, "myobj")):
                if (value is not None):
                  tmp = None
                  # if provided value is not correct type, produce error
                  if (!$cast(tmp, value)):
                    # error
                  else:
                    if(clone)
                      self.myobj = tmp.clone()
                    else
                      self.myobj = tmp
                else:
                  myobj = None  # value is None, so simply assign None to myobj
              end
            ...

        Although the object designer implements these methods to provide outside
        access to one or more properties, they are intended for internal use (e.g.,
        for command-line debugging and auto-configuration) and should not be called
        directly by the user.

        Args:
            field_name (str): Variable to set
            value: Value for the variable
            clone (bool):
            recurse (bool):
        """
        # cc = None  # uvm_object cc
        UVMObject._m_uvm_status_container.cycle_check.clear()
        UVMObject._m_uvm_status_container.m_uvm_cycle_scopes.clear()

        if clone and (value is not None):
            cc = value.clone()
            if cc is not None:
                cc.set_name(field_name)
            value = cc

        UVMObject._m_uvm_status_container.status = False
        UVMObject._m_uvm_status_container.object = value
        UVMObject._m_uvm_status_container.clone = clone

        self._m_uvm_field_automation(None, UVM_SETOBJ, field_name)

        if UVMObject._m_uvm_status_container.warning and not UVMObject._m_uvm_status_container.status:
            uvm_report_error("NOMTC", sv.sformatf("did not find a match for field %s", field_name), UVM_NONE)

        UVMObject._m_uvm_status_container.cycle_check.clear()


    #  //---------------------------------------------------------------------------
    #  //                 **** Internal Methods and Properties ***
    #  //                           Do not use directly
    #  //---------------------------------------------------------------------------
    #

    #  extern local function void m_pack        (inout uvm_packer packer)
    def m_pack(self, packer) -> Any:
        if packer is not None:
            UVMObject._m_uvm_status_container.packer = packer
        else:
            from .uvm_global_vars import uvm_default_packer
            UVMObject._m_uvm_status_container.packer = uvm_default_packer
        packer = UVMObject._m_uvm_status_container.packer

        packer.reset()
        packer.scope.down(self.get_name())

        self._m_uvm_field_automation(None, UVM_PACK, "")
        self.do_pack(packer)
        packer.set_packed_size()
        packer.scope.up()
        return packer


    #  extern local function void m_unpack_pre  (inout uvm_packer packer)
    def m_unpack_pre(self, packer) -> Any:
        if packer is not None:
            UVMObject._m_uvm_status_container.packer = packer
        else:
            from .uvm_global_vars import uvm_default_packer
            UVMObject._m_uvm_status_container.packer = uvm_default_packer
        packer = UVMObject._m_uvm_status_container.packer
        packer.reset()
        return packer

    #  extern local function void m_unpack_post (uvm_packer packer)
    def m_unpack_post(self, packer) -> None:
        provided_size = packer.get_packed_size()
        # Put this object into the hierarchy
        packer.scope.down(self.get_name())

        self._m_uvm_field_automation(None, UVM_UNPACK, "")
        self.do_unpack(packer)
        # Scope back up before leaving
        packer.scope.up()

        if packer.get_packed_size() != provided_size:
            uvm_report_warning("BDUNPK", sv.sformatf(
                "Unpack operation unsuccessful: unpacked %0d bits from a total of %0d bits",
                packer.get_packed_size(), provided_size), UVM_NONE)



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2014 Semifore
#// Copyright 2010-2017 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2014-2015 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
from enum import Enum, auto
from uvm.macros.uvm_message_defines import uvm_error
import cocotb

#// File -- NODOCS -- TLM2 Types
#typedef class uvm_time
#
#// Enum -- NODOCS -- uvm_tlm_phase_e
#//
#// Nonblocking transport synchronization state values between
#// an initiator and a target.
#//
#//  UNINITIALIZED_PHASE  - Defaults for constructor
#//  BEGIN_REQ            - Beginning of request phase
#//  END_REQ              - End of request phase
#//  BEGIN_RESP           - Beginning of response phase
#//  END_RESP             - End of response phase
#

class uvm_tlm_phase_e(Enum):
    UNINITIALIZED_PHASE = auto()
    BEGIN_REQ = auto()
    END_REQ = auto()
    BEGIN_RESP = auto()
    END_RESP = auto()


#// Enum -- NODOCS -- uvm_tlm_sync_e
#//
#// Pre-defined phase state values for the nonblocking transport
#// Base Protocol between an initiator and a target.
#//
#// UVM_TLM_ACCEPTED      - Transaction has been accepted
#// UVM_TLM_UPDATED       - Transaction has been modified
#// UVM_TLM_COMPLETED     - Execution of transaction is complete

class uvm_tlm_sync_e(Enum):
    UVM_TLM_ACCEPTED = auto()
    UVM_TLM_UPDATED = auto()
    UVM_TLM_COMPLETED = auto()

#// Defines Not-Yet-Implemented TLM tasks
UVM_TLM_TASK_ERROR="TLM-2 interface task not implemented"

#// Defines Not-Yet-Implemented TLM functions
UVM_TLM_FUNCTION_ERROR="TLM-2 interface function not implemented"

#//
#// Class -- NODOCS -- uvm_tlm_if
#//
#// Base class type to define the transport functions.
#//
#//  -  <nb_transport_fw>
#//
#//  - <nb_transport_bw>
#//
#//  - <b_transport>
#//

class UVMTLMIf():

    #//----------------------------------------------------------------------
    #// Group: tlm transport methods
    #//
    #// Each of the interface methods take a handle to the transaction to be
    #// transported and a reference argument for the delay. In addition, the
    #// nonblocking interfaces take a reference argument for the phase.
    #//

    #//----------------------------------------------------------------------
    #// Function: nb_transport_fw
    #//
    #// Forward path call.
    #// The first call to this method for a transaction marks the initial timing point.
    #// Every call to this method may mark a timing point in the execution of the
    #// transaction. The timing annotation argument allows the timing points
    #// to be offset from the simulation times at which the forward path is used.
    #// The final timing point of a transaction may be marked by a call
    #// to <nb_transport_bw> or a return from this or subsequent call to
    #// nb_transport_fw.
    #//
    #// See <TLM2 Interfaces, Ports, Exports and Transport Interfaces Subset>
    #// for more details on the semantics and rules of the nonblocking
    #// transport interface.
    def nb_transport_fw(self, t, p, delay):
        uvm_error("nb_transport_fw", UVM_TLM_FUNCTION_ERROR)
        return uvm_tlm_sync_e.UVM_TLM_ACCEPTED

    #// Function: nb_transport_bw
    #//
    #// Implementation of the backward path.
    #// This function MUST be implemented in the INITIATOR component class.
    #//
    #// Every call to this method may mark a timing point, including the final
    #// timing point, in the execution of the transaction.
    #// The timing annotation argument allows the timing point
    #// to be offset from the simulation times at which the backward path is used.
    #// The final timing point of a transaction may be marked by a call
    #// to <nb_transport_fw> or a return from this or subsequent call to
    #// nb_transport_bw.
    #//
    #// See <TLM2 Interfaces, Ports, Exports and Transport Interfaces Subset>
    #// for more details on the semantics and rules of the nonblocking
    #// transport interface.
    #//
    #// Example:
    #//
    #//| class master extends uvm_component;
    #//     uvm_tlm_nb_initiator_socket #(trans, uvm_tlm_phase_e, this_t) initiator_socket;
    #//|    ...
    #//|    function void build_phase(uvm_phase phase);
    #//        initiator_socket = new("initiator_socket", this, this);
    #//|    endfunction
    #//|
    #//|    function uvm_tlm_sync_e nb_transport_bw(ref trans t,
    #//|                                   ref uvm_tlm_phase_e p,
    #//|                                   input uvm_tlm_time delay);
    #//|        transaction = t;
    #//|        state = p;
    #//|        return UVM_TLM_ACCEPTED;
    #//|    endfunction
    #//|
    #//|    ...
    #//| endclass
    def nb_transport_bw(self, t, p, delay):
        uvm_error("nb_transport_bw", UVM_TLM_FUNCTION_ERROR)
        return uvm_tlm_sync_e.UVM_TLM_ACCEPTED

    #// Function: b_transport
    #//
    #// Execute a blocking transaction. Once this method returns,
    #// the transaction is assumed to have been executed. Whether
    #// that execution is successful or not must be indicated by the
    #// transaction itself.
    #//
    #// The callee may modify or update the transaction object, subject
    #// to any constraints imposed by the transaction class. The
    #// initiator may re-use a transaction object from one call to
    #// the next and across calls to b_transport().
    #//
    #// The call to b_transport shall mark the first timing point of the
    #// transaction. The return from b_transport shall mark the final
    #// timing point of the transaction. The timing annotation argument
    #// allows the timing points to be offset from the simulation times
    #// at which the task call and return are executed.
    @cocotb.coroutine
    def b_transport(self, t, delay):
        uvm_error("b_transport", UVM_TLM_TASK_ERROR)




# from cocotb.utils import get_sim_time

from .uvm_object import UVMObject
from .uvm_object_globals import (UVM_BIN, UVM_COUNT, UVM_DEC, UVM_DISPLAY, UVM_ERROR, UVM_EXIT,
                                 UVM_FATAL, UVM_INFO, UVM_LOG, UVM_LOW, UVM_MEDIUM, UVM_NONE,
                                 UVM_NO_ACTION, UVM_RM_RECORD, UVM_SEVERITY_LEVELS, UVM_STOP,
                                 UVM_WARNING)
from .uvm_globals import uvm_report_info, uvm_sim_time
from .uvm_pool import UVMPool
from .uvm_global_vars import uvm_default_printer
from .sv import sv
from ..macros.uvm_message_defines import uvm_info
from .uvm_tr_database import UVMTrDatabase, UVMTextTrDatabase
from typing import List, Any, Callable



def ename(sever) -> str:
    """
    Converts given severity level into string.

    Args:
        sever (int): Severity level.
    Returns:
        str: Severity as string.
    Raises:
        Exception
    """
    if isinstance(sever, str):
        raise Exception("str was given to ename(). Expected int. Got: " + sever)
    if sever == UVM_INFO:
        return "UVM_INFO"
    if sever == UVM_ERROR:
        return "UVM_ERROR"
    if sever == UVM_FATAL:
        return "UVM_FATAL"
    if sever == UVM_WARNING:
        return "UVM_WARNING"
    return "UNKNOWN_SEVERITY: {}".format(sever)



class UVMReportServer(UVMObject):
    """
    UVMReportServer is a global server that processes all of the reports
    generated by a uvm_report_handler.

    The `UVMReportServer` is an abstract class which declares many of its methods
    as ~pure virtual~.  The UVM uses the <uvm_default_report_server> class
    as its default report server implementation.
    """

    def __init__(self, name="base"):
        UVMObject.__init__(self, name)
        self.m_quit_count = 0
        self.m_max_quit_count = 0
        self.max_quit_overridable = True
        self.m_severity_count = UVMPool()
        self.m_id_count = UVMPool()

        self.enable_report_id_count_summary = True
        self.record_all_messages = False
        self.show_verbosity = False
        self.show_terminator = False

        self.m_message_db: UVMTrDatabase = UVMTextTrDatabase()  # uvm_tr_database
        self.m_streams = {}
        self.reset_quit_count()
        self.reset_severity_counts()
        self.set_max_quit_count(0)
        self.print_on_closed_file = True
        self.logger = print  # By default, use print to emit the messages

    def get_type_name(self) -> str:
        return "uvm_report_server"

    @classmethod
    def set_server(cls, server):
        cs = get_cs()
        #server.copy(cs.get_report_server())
        cs.set_report_server(server)

    @classmethod
    def get_server(cls) -> 'UVMReportServer':
        cs = get_cs()
        return cs.get_report_server()

    def set_logger(self, logger: Callable):
        """
        Sets the logger function used to print the messages. Default is python
        built-in print.

        logger (func): Logging function to use.
        """
        self.logger = logger

    # Function: print
    #
    # The uvm_report_server implements the `UVMObject.do_print()` such that
    # ~print~ method provides UVM printer formatted output
    # of the current configuration. A snippet of example output is shown here::
    #
    #  uvm_report_server                 uvm_report_server  -     @13
    #    quit_count                      int                32    'd0
    #    max_quit_count                  int                32    'd5
    #    max_quit_overridable            bit                1     'b1
    #    severity_count                  severity counts    4     -
    #      [UVM_INFO]                    integral           32    'd4
    #      [UVM_WARNING]                 integral           32    'd2
    #      [UVM_ERROR]                   integral           32    'd50
    #      [UVM_FATAL]                   integral           32    'd10
    #    id_count                        id counts          4     -
    #      [ID1]                         integral           32    'd1
    #      [ID2]                         integral           32    'd2
    #      [RNTST]                       integral           32    'd1
    #    enable_report_id_count_summary  bit                1     'b1
    #    record_all_messages             bit                1     `b0
    #    show_verbosity                  bit                1     `b0
    #    show_terminator                 bit                1     `b0

    def do_print(self, printer):
        """
        Print to show report server state

        Args:
            printer (UVMPrinter):
        """
        l_severity_count_index = 0
        l_id_count_index = ""

        printer.print_int("quit_count", self.m_quit_count, sv.bits(self.m_quit_count), UVM_DEC,
            ".", "int")
        printer.print_int("max_quit_count", self.m_max_quit_count,
            sv.bits(self.m_max_quit_count), UVM_DEC, ".", "int")
        printer.print_int("max_quit_overridable", self.max_quit_overridable,
            sv.bits(self.max_quit_overridable), UVM_BIN, ".", "bit")

        if self.m_severity_count.has_first():
            l_severity_count_index = self.m_severity_count.first()
            sev_count = self.m_severity_count.num()
            printer.print_array_header("severity_count",sev_count,"severity counts")
            ok = True
            while ok:
                printer.print_int("[{}]".format(ename(l_severity_count_index)),
                    self.m_severity_count[l_severity_count_index], 32, UVM_DEC)
                ok = self.m_severity_count.has_next()
                if ok:
                    l_severity_count_index = self.m_severity_count.next()
            printer.print_array_footer()

        if self.m_id_count.has_first():
            l_id_count_index = self.m_id_count.first()
            printer.print_array_header("id_count",self.m_id_count.num(),"id counts")
            ok = True
            while ok:
                printer.print_int("[{}]".format(l_id_count_index),
                    self.m_id_count[l_id_count_index], 32, UVM_DEC)
                ok = self.m_id_count.has_next()
                if ok:
                    l_id_count_index = self.m_id_count.next()
            printer.print_array_footer()

        printer.print_int("enable_report_id_count_summary", self.enable_report_id_count_summary,
            sv.bits(self.enable_report_id_count_summary), UVM_BIN, ".", "bit")
        printer.print_int("record_all_messages", self.record_all_messages,
            sv.bits(self.record_all_messages), UVM_BIN, ".", "bit")
        printer.print_int("show_verbosity", self.show_verbosity,
            sv.bits(self.show_verbosity), UVM_BIN, ".", "bit")
        printer.print_int("show_terminator", self.show_terminator,
            sv.bits(self.show_terminator), UVM_BIN, ".", "bit")

    #----------------------------------------------------------------------------
    # Group: Quit Count
    #----------------------------------------------------------------------------

    def get_max_quit_count(self):
        """
        Get the maximum number of COUNT actions that can be tolerated
        before a UVM_EXIT action is taken. The default is 0, which specifies
        no maximum.

        Returns:
            int: Max quit count allowed for the server.
        """
        return self.m_max_quit_count

    # Function: set_max_quit_count
    #
    # Get or set the maximum number of COUNT actions that can be tolerated
    # before a UVM_EXIT action is taken. The default is 0, which specifies
    # no maximum.
    def set_max_quit_count(self, count, overridable=True):
        if self.max_quit_overridable is False:
            uvm_report_info("NOMAXQUITOVR",
                "The max quit count setting of {} is not overridable to {} due to a previous setting."
                .format(self.m_max_quit_count, count), UVM_NONE)
            return
        self.max_quit_overridable = overridable
        if count < 0:
            self.m_max_quit_count = 0
        else:
            self.m_max_quit_count = count

    def get_quit_count(self):
        """
        Function: get_quit_count

        Returns:
            int: Quit count set for this report server.
        """
        return self.m_quit_count

    def set_quit_count(self, quit_count):
        """
        Args:
            quit_count (int): Desired quit count for this server.
        """
        if quit_count < 0:
            self.m_quit_count = 0
        else:
            self.m_quit_count = quit_count

    def incr_quit_count(self):
        """
        Increment quit count by one.
        """
        self.m_quit_count += 1

    def reset_quit_count(self):
        """
        Set, get, increment, or reset to 0 the quit count, i.e., the number of
        `UVM_COUNT` actions issued.
        """
        self.m_quit_count = 0

    def is_quit_count_reached(self):
        """
        If is_quit_count_reached returns 1, then the quit counter has reached
        the maximum.

        Returns:
            bool: True is maximum quit count reached, False otherwise.
        """
        return self.m_quit_count >= self.m_max_quit_count

    #----------------------------------------------------------------------------
    # Group: Severity Count
    #----------------------------------------------------------------------------


    def get_severity_count(self, severity):
        """
        Returns number of messages reported for given severity.

        Args:
            severity (int): Severity level
        Returns:
            int: Number of messages reported for given severity.
        """
        if self.m_severity_count.exists(severity):
            return self.m_severity_count.get(severity)
        return 0

    # Function: set_severity_count
    def set_severity_count(self, severity, count):
        val = count
        if count < 0:
            val = 0
        self.m_severity_count.add(severity, val)

    # Function: incr_severity_count
    def incr_severity_count(self, severity):
        if self.m_severity_count.exists(severity):
            new_count = self.m_severity_count.get(severity) + 1
            self.m_severity_count.add(severity, new_count)
        else:
            self.m_severity_count.add(severity, 1)

    def reset_severity_counts(self):
        """
        Function: reset_severity_counts

        Set, get, or increment the counter for the given severity, or reset
        all severity counters to 0.
        """
        for s in UVM_SEVERITY_LEVELS:
            self.m_severity_count.add(s, 0)

    #----------------------------------------------------------------------------
    # Group: id Count
    #----------------------------------------------------------------------------

    # Function: get_id_count
    def get_id_count(self, id):
        if self.m_id_count.exists(id):
            return self.m_id_count.get(id)
        return 0

    # Function: set_id_count
    def set_id_count(self, id, count):
        val = count
        if count < 0:
            val = 0
        self.m_id_count.add(id, val)

    # Function: incr_id_count
    #
    # Set, get, or increment the counter for reports with the given id.
    def incr_id_count(self, id):
        if self.m_id_count.exists(id):
            val = self.m_id_count.get(id)
            self.m_id_count.add(id, val + 1)
        else:
            self.m_id_count.add(id, 1)

    #----------------------------------------------------------------------------
    # Group: message recording
    #
    # The ~uvm_default_report_server~ will record messages into the message
    # database, using one transaction per message, and one stream per report
    # object/handler pair.
    #
    #----------------------------------------------------------------------------

    def set_message_database(self, database: UVMTrDatabase):
        """
        Function: set_message_database
        sets the `UVMTrDatabase` used for recording messages

        Args:
            database (UVMTrDatabase):
        """
        self.m_message_db = database

    def get_message_database(self) -> UVMTrDatabase:
        """
        Function: get_message_database
        returns the `uvm_tr_database` used for recording messages

        Returns:
            UVMTrDatabase: Message database.
        """
        return self.m_message_db

    def get_severity_set(self, q: List[Any]) -> None:
        while self.m_severity_count.has_next():
            q.append(self.m_severity_count.next())

    def get_id_set(self, q: List[Any]) -> None:
        while self.m_id_count.has_next():
            q.append(self.m_id_count.next())

    def f_display(self, file, _str) -> None:
        """
        This method sends string severity to the command line if file is 0 and to
        the file(s) specified by file if it is not 0.

        Args:
            file:
            _str:
        """
        if file == 0:
            self.logger(_str)
            # print(_str)
        else:
            if not file.closed:
                file.write(_str + "\n")
            else:
                if self.print_on_closed_file:
                    self.logger('UVM_WARNING. File already closed for msg ' + _str)

    def process_report_message(self, report_message) -> None:
        l_report_handler = report_message.get_report_handler()
        #process p = process::self()
        report_ok = True

        # Set the report server for this message
        report_message.set_report_server(self)

        if report_ok is True:
            from .uvm_report_catcher import UVMReportCatcher
            report_ok = UVMReportCatcher.process_all_report_catchers(report_message)

        if report_message.get_action() == UVM_NO_ACTION:
            report_ok = False

        if report_ok:
            m = ""
            cs = get_cs()
            # give the global server a chance to intercept the calls
            svr = cs.get_report_server()

            # no need to compose when neither UVM_DISPLAY nor UVM_LOG is set
            if report_message.get_action() & (UVM_LOG | UVM_DISPLAY):
                m = svr.compose_report_message(report_message)
            svr.execute_report_message(report_message, m)

    #----------------------------------------------------------------------------
    # Group: Message Processing
    #----------------------------------------------------------------------------


    def execute_report_message(self, report_message, composed_message) -> None:
        """
        Processes the provided message per the actions contained within.

        Expert users can overload this method to customize action processing.

        Args:
            report_message (UVMReportMessage):
            composed_message (str): Formatted message string
        """
        #process p = process::self()

        # Update counts
        self.incr_severity_count(report_message.get_severity())
        self.incr_id_count(report_message.get_id())

        if self.record_all_messages is True:
            report_message.set_action(report_message.get_action() | UVM_RM_RECORD)

        # UVM_RM_RECORD action
        if report_message.get_action() & UVM_RM_RECORD:
            stream = None  # uvm_tr_stream stream
            ro = report_message.get_report_object()
            rh = report_message.get_report_handler()

            # Check for pre-existing stream TODO
            #if (m_streams.exists(ro.get_name()) && (m_streams[ro.get_name()].exists(rh.get_name())))
            #stream = m_streams[ro.get_name()][rh.get_name()]

            # If no pre-existing stream (or for some reason pre-existing stream was ~null~)
            if stream is None:
                # Grab the database
                db = self.get_message_database()

                # If database is ~null~, use the default database
                if db is None:
                    cs = get_cs()
                    db = cs.get_default_tr_database()

                if db is not None:
                    # Open the stream.    Name=report object name, scope=report handler name, type=MESSAGES
                    stream = db.open_stream(ro.get_name(), rh.get_name(), "MESSAGES")
                    # Save off the openned stream
                    self.m_streams[ro.get_name()][rh.get_name()] = stream
            if stream is not None:
                recorder = stream.open_recorder(report_message.get_name(), None,report_message.get_type_name())
                if recorder is not None:
                    report_message.record(recorder)
                    recorder.free()

        # DISPLAY action
        if report_message.get_action() & UVM_DISPLAY:
            self.logger(composed_message)

        # LOG action
        # if log is set we need to send to the file but not resend to the
        # display. So, we need to mask off stdout for an mcd or we need
        # to ignore the stdout file handle for a file handle.
        if report_message.get_action() & UVM_LOG:
            if report_message.get_file() == 0 or report_message.get_file() != 0x80000001: #ignore stdout handle
                tmp_file = report_message.get_file()
                #if report_message.get_file() & 0x80000000 == 0: # is an mcd so mask off stdout
                #    tmp_file = report_message.get_file() & 0xfffffffe
                self.f_display(tmp_file, composed_message)

        # Process the UVM_COUNT action
        if report_message.get_action() & UVM_COUNT:
            if self.get_max_quit_count() != 0:
                self.incr_quit_count()
                # If quit count is reached, add the UVM_EXIT action.
                if self.is_quit_count_reached():
                    report_message.set_action(report_message.get_action() | UVM_EXIT)

        # Process the UVM_EXIT action
        if report_message.get_action() & UVM_EXIT:
            cs = get_cs()
            l_root = cs.get_root()
            l_root.die()

        # Process the UVM_STOP action
        if report_message.get_action() & UVM_STOP:
            raise Exception("$stop from uvm_report_server, msg: " +
                    report_message.sprint())

    def compose_report_message(self, report_message, report_object_name="") -> str:
        """
        Constructs the actual string sent to the file or command line
        from the severity, component name, report id, and the message itself.

        Expert users can overload this method to customize report formatting.

        Args:
            report_message (UVMReportMessage):
            report_object_name (UVMReportObject):
        Returns:
            str: Composed message as string.
        """
        sev_string = ""
        l_severity = UVM_INFO
        l_verbosity = UVM_MEDIUM
        filename_line_string = ""
        time_str = ""
        line_str = ""
        context_str = ""
        verbosity_str = ""
        terminator_str = ""
        msg_body_str = ""
        el_container = None
        prefix = ""
        l_report_handler = None

        l_severity = report_message.get_severity()
        sev_string = ename(l_severity)

        if report_message.get_filename() != "":
            line_str = str(report_message.get_line())
            filename_line_string = report_message.get_filename() + "(" + line_str + ") "

        # Make definable in terms of units.
        #$swrite(time_str, "%0t", $time)
        #time_str = get_sim_time('ns') TODO
        time_str = str(uvm_sim_time('NS')) + 'NS'

        if report_message.get_context() != "":
            context_str = "@@" + report_message.get_context()

        if self.show_verbosity is True:
            verb = report_message.get_verbosity()
            verbosity_str = "(" + verb + ")"

        if self.show_terminator is True:
            terminator_str = " -" + sev_string

        el_container = report_message.get_element_container()
        if el_container.size() == 0:
            msg_body_str = report_message.get_message()
        else:
            prefix = uvm_default_printer.knobs.prefix
            uvm_default_printer.knobs.prefix = " +"
            msg_body_str = report_message.get_message() + "\n" + el_container.sprint()
            uvm_default_printer.knobs.prefix = prefix

        if report_object_name == "":
            l_report_handler = report_message.get_report_handler()
            if l_report_handler is not None:
                report_object_name = l_report_handler.get_full_name()
            else:
                report_object_name = "NO_REPORT_OBJECT"

        result = (sev_string + verbosity_str + " " + filename_line_string
            + "@ " + time_str + ": " + report_object_name + context_str
            + " [" + report_message.get_id() + "] " + msg_body_str + terminator_str)
        return result


    def report_summarize(self, file=0) -> None:
        """
        Outputs statistical information on the reports issued by this central
        report server. This information will be sent to the command line if
        ~file~ is 0, or to the file descriptor ~file~ if it is not 0.

        The `UVMRoot.run_test` method in `UVMRoot` calls this method.
        """
        rpt = self.get_summary_string()
        uvm_info("UVM/REPORT/SERVER", rpt, UVM_LOW)

    def get_summary_string(self) -> str:
        """
        Returns the statistical information on the reports issued by this central report
        server as multi-line string.

        Returns:
            str: End of simulation summary.
        """
        id = ""
        q = []

        from .uvm_report_catcher import UVMReportCatcher
        UVMReportCatcher.summarize()
        q.append("\n--- UVM Report Summary ---\n\n")

        if self.m_max_quit_count != 0:
            if self.m_quit_count >= self.m_max_quit_count:
                q.append("Quit count reached!\n")
            q.append("Quit count : {} of {}\n".format(self.m_quit_count,
                self.m_max_quit_count))

        q.append("** Report counts by severity\n")
        for s in self.m_severity_count.keys():
            q.append("{} : {}\n".format(ename(s), self.m_severity_count.get(s)))

        if self.enable_report_id_count_summary is True:
            q.append("** Report counts by id\n")
            for id in self.m_id_count.keys():
                q.append("[{}] {}\n".format(id, self.m_id_count.get(id)))
        return "".join(q)


def get_cs():
    from .uvm_coreservice import UVMCoreService
    return UVMCoreService.get()




import unittest

from uvm.base.uvm_callback import (UVMCallback, UVMCallbacks,
    UVMCallbackIter, UVMTypedCallbacks, UVMCallbacksBase)

from uvm.base.uvm_component import UVMComponent

from uvm.uvm_unit import MockObj, MockCb
from uvm.macros import (uvm_register_cb, uvm_do_callbacks)

UVMCallbacksBase.m_tracing = True


class TestUVMCallback(unittest.TestCase):

    def test_typed_callbacks(self):
        typed_cb = UVMTypedCallbacks.m_initialize()
        self.assertEqual(isinstance(typed_cb, UVMTypedCallbacks), True)
        typed_cb2 = UVMTypedCallbacks.m_initialize()
        self.assertEqual(typed_cb, typed_cb2)

    def test_callback_mode(self):
        cb = UVMCallback('my_cb')
        cb.callback_mode(1)
        self.assertEqual(cb.is_enabled(), True)

    def test_type_name(self):
        cb = UVMCallback('my_cb')
        self.assertEqual(cb.get_type_name(), 'uvm_callback')

    def test_callback_iter(self):
        obj = MockObj("parent", None)
        my_iter = UVMCallbackIter(obj)
        cb = my_iter.first()
        self.assertEqual(cb, None)
        cb = my_iter.last()
        self.assertEqual(cb, None)
        cb_obj = MockCb("cb_obj")
        UVMCallbacks.add(obj, cb_obj)
        cb = my_iter.first()
        self.assertEqual(cb, cb_obj)
        cb = my_iter.last()
        self.assertEqual(cb, cb_obj)

    def test_uvm_callbacks(self):
        parent_obj = MockObj("parent", None)
        cb_obj = MockCb("cb_obj")
        my_iter = UVMCallbackIter(parent_obj)
        UVMCallbacks.add(parent_obj, cb_obj)

        cb = UVMCallbacks.get_first(my_iter, parent_obj)
        self.assertEqual(cb, cb_obj)

        cb2 = my_iter.first()
        self.assertEqual(cb2, cb_obj)

        class MyCallback(UVMCallback):
            def get_name(self):
                return "my_cb"
            def do_it(self):
                self.called = True
            def get_type_name(self):
                return "MyCallback"

        class MyCallbackWrong(UVMCallback):
            def do_it(self):
                raise Exception('Dont call this')
            def get_type_name(self):
                return "MyCallbackWrong"

        class MyClass(UVMComponent):
            def do_it(self):
                uvm_do_callbacks(self, MyCallback, 'do_it')
        uvm_register_cb(MyClass, MyCallback)
        uvm_register_cb(MyClass, MyCallbackWrong)

        my_class = MyClass('my_class_obj', None)
        my_cb = MyCallback()
        UVMCallbacks.add(my_class, my_cb)
        UVMCallbacks.add(my_class, MyCallbackWrong())
        my_class.do_it()
        self.assertTrue(my_cb.called)




if __name__ == '__main__':
    unittest.main()



#
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

# About: factory
# This example will illustrate the usage of uvm_factory methods.

# To get more details about the factory related methods, check the file:
#   - uvm/src/base/uvm_factory.py


from uvm.macros import *
from uvm import *
from packet_pkg import packet
from gen_pkg import gen
from env_pkg import env


#
#  env e;  //you need to use something from the package to have the
#          //factory registration occur.
#


class mygen(gen):
    def __init__(self, name, parent):
        super().__init__(name, parent)


    def get_packet(self):
        uvm_info("PKTGEN", sv.sformatf("Getting a packet from %s (%s)",
            self.get_full_name(), self.get_type_name()), UVM_MEDIUM)
        return super().get_packet()


# Use the macro in a class to implement factory registration along with other
# utilities (create, get_type_name). To just do factory registration, use the
# macro `uvm_object_registry(mygen,"mygen")
uvm_component_utils(mygen)

class mypacket(packet):
    #    constraint ct10 { addr >= 0  and  addr <= 10; }

    def __init__(self, name="mypacket"):
        super().__init__(name)
        self.my_packet_prop = 123

# Use the macro in after the class to implement factory registration along with other
# utilities (create, get_type_name).
uvm_object_utils(mypacket)


@cocotb.test()
async def module_top(dut):
    cs_ = UVMCoreService.get()
    factory = cs_.get_factory()
    gen.type_id.set_inst_override(mygen.get_type(), "uvm_test_top.gen1")
    packet.type_id.set_type_override(mypacket.get_type())
    factory.print_factory(1)

    # If a string is used to run_test, run_test will used the string based factory
    # create method to create an object of the desired type.
    await run_test("env")



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.base.uvm_callback import *
import cocotb
from uvm.comps.uvm_driver import *
from uvm.macros import *
from cocotb.triggers import *

class apb_master_cbs(UVMCallback):
    
    async def trans_received(self, xactor, cycle):
        await Timer(0)

    
    async def trans_executed(self, xactor, cycle):
        await Timer(0)


class apb_master(UVMDriver):  # (apb_rw)

    def __init__(self, name, parent=None):
        super().__init__(name,parent)
        self.trig = None  # event
        self.sigs = None  # apb_vif
        self.cfg = None  # apb_config
        self.tr = None  # apb_rw
        #   endfunction


    def build_phase(self, phase):
        agent = None
        if (sv.cast(agent, self.get_parent(), apb_agent) and agent is not None):
            self.sigs = agent.vif
        else:
           if (not UVMConfigDb.get(self, "", "vif", sigs)):
              uvm_fatal("APB/DRV/NOVIF", "No virtual interface specified for self driver instance")
    #   endfunction

    #   def run_phase(self,uvm_phase phase)
    #      super().run_phase(phase)
    #
    #      self.sigs.mck.psel    <=  0
    #      self.sigs.mck.penable <=  0
    #
    #      while True:
    #         apb_rw tr
    #         yield Edge(self.sigs.mck)
    #
    #         seq_item_port.get_next_item(tr)
    #
    #         // TODO: QUESTA issue with hier ref to sequence via modport; need workaround?
    #	      if (!self.sigs.mck.triggered)
    #             yield Edge(self.sigs.mck)
    #
    #         self.trans_received(tr)
    #         `uvm_do_callbacks(apb_master,apb_master_cbs,trans_received(self,tr))
    #
    #	      case (tr.kind)
    #          apb_rw::READ:  self.read(tr.addr, tr.data);
    #          apb_rw::WRITE: self.write(tr.addr, tr.data)
    #         endcase
    #
    #         self.trans_executed(tr)
    #         `uvm_do_callbacks(apb_master,apb_master_cbs,trans_executed(self,tr))
    #
    #         seq_item_port.item_done()
    #	      ->trig
    #      end
    #   endtask: run_phase


    #   def read(self,input  bit   [31:0] addr,
    #                               output logic [31:0] data)
    #
    #      self.sigs.mck.paddr   <= addr
    #      self.sigs.mck.pwrite  <=  0
    #      self.sigs.mck.psel    <=  1
    #      yield Edge(self.sigs.mck)
    #      self.sigs.mck.penable <=  1
    #      yield Edge(self.sigs.mck)
    #      data = self.sigs.mck.prdata
    #      self.sigs.mck.psel    <=  0
    #      self.sigs.mck.penable <=  0
    #   endtask: read
    #
    #   def write(self,input bit [31:0] addr,
    #                                input bit [31:0] data)
    #      self.sigs.mck.paddr   <= addr
    #      self.sigs.mck.pwdata  <= data
    #      self.sigs.mck.pwrite  <=  1
    #      self.sigs.mck.psel    <=  1
    #      yield Edge(self.sigs.mck)
    #      self.sigs.mck.penable <=  1
    #      yield Edge(self.sigs.mck)
    #      self.sigs.mck.psel    <=  0
    #      self.sigs.mck.penable <=  0
    #   endtask: write
    #
    #   def trans_received(self,apb_rw tr)
    #   endtask
    #
    #   def trans_executed(self,apb_rw tr)
    #   endtask
    #endclass: apb_master
uvm_component_utils(apb_master)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


from uvm.base import *
from uvm.comps import UVMEnv
from uvm.macros import uvm_component_utils
from ubus_example_scoreboard import ubus_example_scoreboard

from ubus_env import ubus_env


class ubus_example_tb(UVMEnv):
    #  // Provide implementations of virtual methods such as get_type_name and create

    def __init__(self, name, parent=None):
        UVMEnv.__init__(self, name, parent)
        self.scoreboard0 = None
        self.ubus0 = None

    #  // build_phase
    def build_phase(self, phase):
        UVMEnv.build_phase(self, phase)

        num_masters = []
        if UVMConfigDb.get(self, "ubus0", "num_masters", num_masters):
            num_masters = num_masters[0]
        else:
            num_masters = 1
        UVMConfigDb.set(self, "ubus0", "num_masters", num_masters)

        num_slaves = []
        if UVMConfigDb.get(self, "ubus0", "num_slaves", num_slaves):
            num_slaves = num_slaves[0]
        else:
            num_slaves = 1
        UVMConfigDb.set(self, "ubus0", "num_slaves", num_slaves)

        self.ubus0 = ubus_env.type_id.create("ubus0", self)
        self.scoreboard0 = ubus_example_scoreboard.type_id.create("scoreboard0",
            self)

    def connect_phase(self, phase):
        # Connect slave0 monitor to scoreboard
        self.ubus0.slaves[0].monitor.item_collected_port.connect(
            self.scoreboard0.item_collected_export)

    def end_of_elaboration_phase(self, phase):
        # Set up slave address map for ubus0 (basic default)
        self.ubus0.set_slave_address_map("slaves[0]", 0, 0xffff)


uvm_component_utils(ubus_example_tb)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.base.uvm_component import *
from uvm.macros import *
from uvm import *

from apb.apb_agent import apb_agent
from reg_model import reg_block_slave
from apb.apb_rw import reg2apb_adapter


class tb_env(UVMComponent):

    #    reg_block_slave model;
    #    apb_agent apb
    #
    def __init__(self, name, parent=None):
        super().__init__(name,parent)
        self.model = None  # type: reg_block_slave
        self.apb = None  # type: apb_agent


    def build_phase(self, phase):
        dut = []
        if UVMConfigDb.get(self, "", "dut", dut):
            dut = dut[0]
        if self.model is None:
            self.model = reg_block_slave.type_id.create("model", self)
            if hasattr(dut, 'NSESS'):
                self.model.nsession = dut.NSESS.value
            elif sv.test_plusargs("NSESS"):
                self.model.nsession = sv.value_plusargs("NSESS")
            else:
                self.model.nsession = 128
            self.model.build()
            self.model.lock_model()

        self.apb = apb_agent.type_id.create("apb", self)


    def connect_phase(self, phase):
        fname = open('uvm_logfile.txt', 'w+')
        self.set_report_default_file_hier(fname)
        self.set_report_severity_action_hier(UVM_INFO, UVM_LOG | UVM_DISPLAY)
        if self.model.get_parent() is None:
            reg2apb = reg2apb_adapter()
            self.model.default_map.set_sequencer(self.apb.sqr, reg2apb)
            self.model.default_map.set_auto_predict(1)



    def end_of_elaboration(self):
        self.model.print_obj()



uvm_component_utils(tb_env)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2008 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2013 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
"""

Title: Register Access Test Sequences

This section defines sequences that test DUT register access via the
available frontdoor and backdoor paths defined in the provided register
model.

"""

from cocotb.triggers import Timer

from .uvm_mem_access_seq import UVMMemAccessSeq
from ..uvm_reg_model import *
from ..uvm_reg_map import UVMRegMap
from ..uvm_reg_sequence import UVMRegSequence
from ...macros import uvm_object_utils, uvm_error, uvm_info, uvm_warning
from ...base import UVMResourceDb, UVM_LOW, uvm_zero_delay
from ...base.uvm_globals import uvm_report_info


#//------------------------------------------------------------------------------
#//
#// Class: UVMRegSingleAccessSeq
#//
#// Verify the accessibility of a register
#// by writing through its default address map
#// then reading it via the backdoor, then reversing the process,
#// making sure that the resulting value matches the mirrored value.
#//
#// If bit-type resource named
#// "NO_REG_TESTS" or "NO_REG_ACCESS_TEST"
#// in the "REG::" namespace
#// matches the full name of the register,
#// the register is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.r0.get_full_name()},
#//|                            "NO_REG_TESTS", 1, this);
#//
#// Registers without an available backdoor or
#// that contain read-only fields only,
#// or fields with unknown access policies
#// cannot be tested.
#//
#// The DUT should be idle and not modify any register during this test.
#//
#//------------------------------------------------------------------------------


class UVMRegSingleAccessSeq(UVMRegSequence):

    def __init__(self, name="UVMRegSingleAccessSeq"):
        super().__init__(name)
        #   // Variable: rg
        #   // The register to be tested
        self.reg = None



    async def body(self):
        maps = []  # uvm_reg_map[$]
        rg = self.rg

        if rg is None:
            uvm_error("UVMRegAccessSeq", "No register specified to run sequence on")
            return


        # Registers with some attributes are not to be tested
        if (UVMResourceDb.get_by_name("REG::" + rg.get_full_name(),
            "NO_REG_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + rg.get_full_name(),
                "NO_REG_ACCESS_TEST", 0) is not None):
            return

        # Can only deal with registers with backdoor access
        if (rg.get_backdoor() is None and not rg.has_hdl_path()):
            uvm_error("UVMRegAccessSeq", ("Register '" + rg.get_full_name()
                + "' does not have a backdoor mechanism available"))
            return


        # Registers may be accessible from multiple physical interfaces (maps)
        rg.get_maps(maps)

        # Cannot test access if register contains RO or OTHER fields

        fields = []  # uvm_reg_field[$]
        rg.get_fields(fields)
        for k in range(len(maps)):
            ro = 0
            for j in range(len(fields)):
                if fields[j].get_access(maps[k]) == "RO":
                    ro += 1

                if not fields[j].is_known_access(maps[k]):
                    uvm_warning("UVMRegAccessSeq", ("Register '" + rg.get_full_name()
                        + "' has field with unknown access type '"
                        + fields[j].get_access(maps[k]) + "', skipping"))
                    return

            if ro == len(fields):
                uvm_warning("UVMRegAccessSeq", ("Register '" + rg.get_full_name()
                    + "' has only RO fields in map " + maps[k].get_full_name() + ", skipping"))
                return



        # Access each register:
        # - Write complement of reset value via front door
        # - Read value via backdoor and compare against mirror
        # - Write reset value via backdoor
        # - Read via front door and compare against mirror

        for j in range(len(maps)):
            status = 0
            v = 0
            exp = 0

            uvm_info("UVMRegAccessSeq", ("Verifying access of register '"
                + rg.get_full_name() + "' in map '" +  maps[j].get_full_name()
                + "' ..."), UVM_LOW)

            v = rg.get()
            status = []
            await rg.write(status, ~v, UVM_FRONTDOOR, maps[j], self)
            status = status[0]

            if status != UVM_IS_OK:
                uvm_error("UVMRegAccessSeq", ("Status was '" + status.name()
                   + "' when writing '" + rg.get_full_name()
                   + "' through map '" + maps[j].get_full_name() + "'"))
            await Timer(1, "NS")  # tpoikela: Original UVM has #1

            status = []
            await rg.mirror(status, UVM_CHECK, UVM_BACKDOOR, UVMRegMap.backdoor(), self)
            status = status[0]
            if status != UVM_IS_OK:
                uvm_error("UVMRegAccessSeq", ("Status was '" + status.name()
                                    + "' when reading reset value of register '"
                                    + rg.get_full_name() +  "' through backdoor"))

            status = []
            await rg.write(status, v, UVM_BACKDOOR, maps[j], self)
            status = status[0]
            if status != UVM_IS_OK:
                uvm_error("UVMRegAccessSeq", ("Status was '" + status.name()
                   + "' when writing '" + rg.get_full_name()
                   + "' through backdoor"))

            status = []
            await rg.mirror(status, UVM_CHECK, UVM_FRONTDOOR, maps[j], self)
            status = status[0]
            if status != UVM_IS_OK:
                uvm_error("UVMRegAccessSeq", ("Status was '" + status.name()
                   + "' when reading reset value of register '"
                   + rg.get_full_name() + "' through map '"
                   + maps[j].get_full_name() + "'"))


uvm_object_utils(UVMRegSingleAccessSeq)


#//------------------------------------------------------------------------------
#//
#// Class: UVMRegAccessSeq
#//
#// Verify the accessibility of all registers in a block
#// by executing the <UVMRegSingleAccessSeq> sequence on
#// every register within it.
#//
#// If bit-type resource named
#// "NO_REG_TESTS" or "NO_REG_ACCESS_TEST"
#// in the "REG::" namespace
#// matches the full name of the block,
#// the block is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.get_full_name(),".*"},
#//|                            "NO_REG_TESTS", 1, this);
#//
#//------------------------------------------------------------------------------


class UVMRegAccessSeq(UVMRegSequence):


    def __init__(self, name="UVMRegAccessSeq"):
        super().__init__(name)
        #   // Variable: model
        #   //
        #   // The block to be tested. Declared in the base class.
        #   //
        #   //| uvm_reg_block model;

        #   // Variable: reg_seq
        #   //
        #   // The sequence used to test one register
        #   //
        self.reg_seq = None



    #   // Task: body
    #   //
    #   // Executes the Register Access sequence.
    #   // Do not call directly. Use seq.start() instead.
    #   //
    async def body(self):
        model = self.model

        if model is None:
            uvm_error("UVMRegAccessSeq", "No register model specified to run sequence on")
            return


        uvm_info("STARTING_SEQ","\n\nStarting " + self.get_name() + " sequence...\n", UVM_LOW)
        self.reg_seq = UVMRegSingleAccessSeq.type_id.create("single_reg_access_seq")
        await self.reset_blk(model)
        model.reset()
        await self.do_block(model)


    #   // Task: do_block
    #   //
    #   // Test all of the registers in a block
    #   //
    async def do_block(self, blk):
        #      uvm_reg regs[$]
        regs = []

        if reg_test_off(blk, "NO_REG_TESTS") or reg_test_off(blk,
                "NO_REG_ACCESS_TEST"):
            return

        # Iterate over all registers, checking accesses
        blk.get_registers(regs, UVM_NO_HIER)
        for i in range(len(regs)):
            # Registers with some attributes are not to be tested
            if reg_test_off(regs[i], "NO_REG_TESTS") or reg_test_off(regs[i],
                    "NO_REG_ACCESS_TEST"):
                continue

            # Can only deal with registers with backdoor access
            if (regs[i].get_backdoor() is None and not regs[i].has_hdl_path()):
                uvm_warning("UVMRegAccessSeq", "Register '" + regs[i].get_full_name()
                       + "' does not have a backdoor mechanism available")
                continue

            self.reg_seq.rg = regs[i]
            await self.reg_seq.start(None,self)

        blks = []  # uvm_reg_block[$]
        blk.get_blocks(blks)
        for i in range(len(blks)):
            await self.do_block(blks[i])


    #   // Task: reset_blk
    #   //
    #   // Reset the DUT that corresponds to the specified block abstraction class.
    #   //
    #   // Currently empty.
    #   // Will rollback the environment's phase to the ~reset~
    #   // phase once the new phasing is available.
    #   //
    #   // In the meantime, the DUT should be reset before executing this
    #   // test sequence or this method should be implemented
    #   // in an extension to reset the DUT.
    #   //
    async def reset_blk(self, blk):
        await uvm_zero_delay()


uvm_object_utils(UVMRegAccessSeq)



class UVMRegMemAccessSeq(UVMRegSequence):  # (uvm_sequence #(uvm_reg_item))
    """
    Class: UVMRegMemAccessSeq

    Verify the accessibility of all registers and memories in a block
    by executing the `UVMRegAccessSeq` and
    `UVMMemAccessSeq` sequence respectively on every register
    and memory within it.

    Blocks and registers with the NO_REG_TESTS or
    the NO_REG_ACCESS_TEST attribute are not verified.
    """


    def __init__(self, name="UVMRegMemAccessSeq"):
        super().__init__(name)



    async def body(self):
        model = self.model

        if model is None:
            uvm_error("UVMRegMemAccessSeq", "Register model handle is None")
            return

        uvm_report_info("STARTING_SEQ",
            "\n\nStarting " + self.get_name() + " sequence...\n", UVM_LOW)

        if reg_test_on(model, "NO_REG_TESTS") and reg_test_on(model,
                "NO_REG_ACCESS_TEST"):
            sub_seq = UVMRegAccessSeq("reg_access_seq")
            await self.reset_blk(model)
            model.reset()
            sub_seq.model = model
            await sub_seq.start(None,self)

        if reg_test_on(model, "NO_MEM_ACCESS_TEST"):
            sub_seq = UVMMemAccessSeq("mem_access_seq")
            await self.reset_blk(model)
            model.reset()
            sub_seq.model = model
            await sub_seq.start(None,self)


    #   // Any additional steps required to reset the block
    #   // and make it accessibl
    async def reset_blk(self, blk):
        await uvm_zero_delay()


uvm_object_utils(UVMRegMemAccessSeq)



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm.base.uvm_component import *
from uvm.macros import *
from usb_xfer import usb_xfer, USB_TLM_HANDSHAKE, USB_TLM_TOKEN, USB_TLM_DATA
from uvm.tlm2 import UVMTLMNbTargetSocket, uvm_tlm_sync_e


class device(UVMComponent):


    def __init__(self, name="device", parent=None):
        super().__init__(name, parent)
        self.sock = UVMTLMNbTargetSocket("sock", self)
        self.data = []


    async def fork_proc(self, xf, dl):
        ph = USB_TLM_HANDSHAKE
        await Timer(100, "NS")
        xf.status = usb_xfer.ACK
        assert(sock.nb_transport_bw(xf, ph, dl) == UVM_TLM_COMPLETED)
        uvm_info("USB/DEV/OUT/DONE", xf.convert2string(),
            UVM_NONE)


    #   // Forward path
    async def nb_transport_fw(self, xfer, ph, delay):
        uvm_info("USB/DEV/FWD", sv.sformatf("%s @%0d: %s",
            int(ph), delay.get_realtime(1),
            xfer.convert2string()), UVM_LOW)
        # This device?
        if xfer.addr != 0x7A:
            return uvm_tlm_sync_e.UVM_TLM_COMPLETED
        
        # Valid endpoint in the device?
        if xfer.endp != 0x5:
            return uvm_tlm_sync_e.UVM_TLM_COMPLETED

        if xfer.kind == usb_xfer.OUT:
            if ph == USB_TLM_TOKEN:
                return uvm_tlm_sync_e.UVM_TLM_ACCEPTED
            elif ph == USB_TLM_DATA:
                data = xfer.data
                xf = xfer
                dl = delay

                # Could complete transfer early here
                #fork: out_ack
                cocotb.fork(self.fork_proc(xf, dl))
                #join_none
                return uvm_tlm_sync_e.UVM_TLM_ACCEPTED

        elif xfer.kind == usb_xfer.IN:
            if ph == USB_TLM_TOKEN:
                # Could return the data early here
                #fork: in_data
                xf = xfer
                dl = delay
                cocotb.fork(self.fork_proc2(xf, dl))
                #join_none
                return uvm_tlm_sync_e.UVM_TLM_ACCEPTED
            elif ph == USB_TLM_HANDSHAKE:
                uvm_info("USB/DEV/IN/DONE", xfer.convert2string(),
                    UVM_NONE)
                return uvm_tlm_sync_e.UVM_TLM_COMPLETED

        return uvm_tlm_sync_e.UVM_TLM_COMPLETED


    async def fork_proc2(self, xf, dl):
        ph = USB_TLM_DATA
        await Timer(150, )
        tdata = [0xAB, 0xCD]
        xf.data = tdata
        self.sock.nb_transport_bw(xf, ph, dl)  # cast to 'void' removed


uvm_component_utils(device)



# ------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010-2011 Synopsys, Inc.
#   Copyright 2013-2014 NVIDIA Corporation
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#
#   uvm-python NOTE: All code ported from SystemVerilog UVM 1.2 to
#   python. Original code structures (including comments)
#   preserved where possible.
# ------------------------------------------------------------------------------

import cocotb
from cocotb.triggers import Edge, ReadWrite, NullTrigger
from cocotb.utils import get_sim_time, simulator

from .uvm_object_globals import (UVM_CALL_HOOK, UVM_COUNT, UVM_DISPLAY, UVM_ERROR, UVM_EXIT,
                                 UVM_FATAL, UVM_INFO, UVM_LOG, UVM_LOW, UVM_MEDIUM, UVM_NONE,
                                 UVM_NO_ACTION, UVM_RM_RECORD, UVM_STOP, UVM_WARNING)
from .sv import uvm_glob_to_re, uvm_re_match
from inspect import getframeinfo, stack

"""
Title: Globals
Group: Simulation Control
"""


async def run_test(test_name="", dut=None):
    """
    Convenience function for uvm_top.run_test(). See `UVMRoot` for more
    information.

    Args:
        test_name (str): Name of the test to run.
        dut: DUT object from cocotb
    """
    cs = get_cs()
    top = cs.get_root()
    await top.run_test(test_name, dut)

#//----------------------------------------------------------------------------
#//
#// Group: Reporting
#//
#//----------------------------------------------------------------------------

#// Function: uvm_get_report_object
#//
def uvm_get_report_object():
    """
    Returns the nearest uvm_report_object when called.
    For the global version, it returns `UVMRoot`.
    """
    from .uvm_coreservice import UVMCoreService
    cs = UVMCoreService.get()
    top = cs.get_root()
    return top


def uvm_report_enabled(verbosity, severity=UVM_INFO, id=""):
    """
    Returns 1 if the configured verbosity in ~uvm_top~ for this
    severity/id is greater than or equal to ~verbosity~ else returns 0.

    See also `UVMReportObject.uvm_report_enabled`.

    Static methods of an extension of `UVMReportObject`, e.g. uvm_component-based
    objects, cannot call `uvm_report_enabled` because the call will resolve to
    the `UVMReportObject.uvm_report_enabled`, which is non-static.
    Static methods cannot call non-static methods of the same class.
    """
    cs = get_cs()
    top = cs.get_root()
    return top.uvm_report_enabled(verbosity,severity,id)


#// Function: uvm_report
#
#function void uvm_report( uvm_severity severity,
#                          string id,
#                          string message,
#                          int verbosity = (severity == uvm_severity'(UVM_ERROR)) ? UVM_LOW :
#                                          (severity == uvm_severity'(UVM_FATAL)) ? UVM_NONE : UVM_MEDIUM,
#                          string filename = "",
#                          int line = 0,
#                          string context_name = "",
#                          bit report_enabled_checked = 0)
#  uvm_root top
#  uvm_coreservice_t cs
#  cs = uvm_coreservice_t::get()
#  top = cs.get_root()
#  top.uvm_report(severity, id, message, verbosity, filename, line, context_name, report_enabled_checked)
#endfunction


def uvm_report_info(id, message, verbosity=UVM_MEDIUM, filename="", line=0,
        context_name="", report_enabled_checked=False):
    if uvm_report_enabled(verbosity, UVM_INFO, id):
        cs = get_cs()
        top = cs.get_root()
        # tpoikela: Do not refactor
        if filename == "":
            caller = getframeinfo(stack()[1][0])
            filename = caller.filename
        if line == 0:
            caller = getframeinfo(stack()[1][0])
            line = caller.lineno
        top.uvm_report_info(id, message, verbosity, filename, line, context_name,
                report_enabled_checked)


def uvm_report_error(id, message, verbosity=UVM_LOW, filename="", line=0,
        context_name="", report_enabled_checked=False):
    if uvm_report_enabled(verbosity, UVM_ERROR, id):
        cs = get_cs()
        top = cs.get_root()
        # tpoikela: Do not refactor
        if filename == "":
            caller = getframeinfo(stack()[1][0])
            filename = caller.filename
        if line == 0:
            caller = getframeinfo(stack()[1][0])
            line = caller.lineno
        top.uvm_report_error(id, message, verbosity, filename, line, context_name,
                report_enabled_checked)


def uvm_report_warning(id, message, verbosity=UVM_LOW, filename="", line=0,
        context_name="", report_enabled_checked=False):
    if uvm_report_enabled(verbosity, UVM_WARNING, id):
        cs = get_cs()
        top = cs.get_root()
        # tpoikela: Do not refactor
        if filename == "":
            caller = getframeinfo(stack()[1][0])
            filename = caller.filename
        if line == 0:
            caller = getframeinfo(stack()[1][0])
            line = caller.lineno
        top.uvm_report_warning(id, message, verbosity, filename, line, context_name,
                report_enabled_checked)


def uvm_report_fatal(id, message, verbosity=UVM_NONE, filename="", line=0,
        context_name="", report_enabled_checked=False):
    """
    These methods, defined in package scope, are convenience functions that
    delegate to the corresponding component methods in ~uvm_top~. They can be
    used in module-based code to use the same reporting mechanism as class-based
    components. See `UVMReportObject for details on the reporting mechanism.

    *Note:* Verbosity is ignored for warnings, errors, and fatals to ensure users
    do not inadvertently filter them out. It remains in the methods for backward
    compatibility.
    """

    if uvm_report_enabled(verbosity, UVM_FATAL, id):
        cs = get_cs()
        top = cs.get_root()
        # tpoikela: Do not refactor
        if filename == "":
            caller = getframeinfo(stack()[1][0])
            filename = caller.filename
        if line == 0:
            caller = getframeinfo(stack()[1][0])
            line = caller.lineno
        top.uvm_report_fatal(id, message, verbosity, filename, line, context_name,
                report_enabled_checked)


def uvm_process_report_message(report_message):
    """
    This method, defined in package scope, is a convenience function that
    delegate to the corresponding component method in ~uvm_top~.
    See `UVMReportObject for details on the reporting mechanism.
    """
    from .uvm_coreservice import UVMCoreService
    cs = UVMCoreService.get()
    top = cs.get_root()
    top.uvm_process_report_message(report_message)


#// TODO merge with uvm_enum_wrapper#(uvm_severity)
def uvm_string_to_severity(sev_str, sev):
    if sev_str == "UVM_INFO":
        return UVM_INFO
    if sev_str == "UVM_WARNING":
        return UVM_WARNING
    if sev_str == "UVM_ERROR":
        return UVM_ERROR
    if sev_str == "UVM_FATAL":
        return UVM_FATAL
    return -1


def uvm_string_to_action(action_str, action):
    #  string actions[$]
    actions = action_str.split('|')  # uvm_split_string(action_str,"|",actions)
    action = 0x0
    for i in range(len(actions)):
        if actions[i] == "UVM_NO_ACTION":
            action |= UVM_NO_ACTION
        if actions[i] == "UVM_DISPLAY":
            action |= UVM_DISPLAY
        if actions[i] == "UVM_LOG":
            action |= UVM_LOG
        if actions[i] == "UVM_COUNT":
            action |= UVM_COUNT
        if actions[i] == "UVM_EXIT":
            action |= UVM_EXIT
        if actions[i] == "UVM_CALL_HOOK":
            action |= UVM_CALL_HOOK
        if actions[i] == "UVM_STOP":
            action |= UVM_STOP
        if actions[i] == "UVM_RM_RECORD":
            action |= UVM_RM_RECORD
    return action

"""
Group: Miscellaneous
"""

def uvm_is_match(expr, _str):
    """
    Returns 1 if the two strings match, 0 otherwise.

    The first string, ~expr~, is a string that may contain '*' and '?'
    characters. A * matches zero or more characters, and ? matches any single
    character. The 2nd argument, ~str~, is the string begin matched against.
    It must not contain any wildcards.
    """
    s = uvm_glob_to_re(expr)
    return uvm_re_match(s, _str) == 0


UVM_LINE_WIDTH = 120
UVM_NUM_LINES = 120
UVM_SMALL_STRING = UVM_LINE_WIDTH*8-1
UVM_LARGE_STRING = UVM_LINE_WIDTH*UVM_NUM_LINES*8-1

#//----------------------------------------------------------------------------
#//
#// Function: uvm_string_to_bits
#//
#// Converts an input string to its bit-vector equivalent. Max bit-vector
#// length is approximately 14000 characters.
#//----------------------------------------------------------------------------


def uvm_string_to_bits(string: str) -> int:
    bytearr = string.encode()
    result = 0
    shift = 0
    for bb in bytearr:
        result = (result << shift) | bb
        shift += 8
    return result


#//----------------------------------------------------------------------------
#//
#// Function: uvm_bits_to_string
#//
#// Converts an input bit-vector to its string equivalent. Max bit-vector
#// length is approximately 14000 characters.
#//----------------------------------------------------------------------------

#function string uvm_bits_to_string(logic [UVM_LARGE_STRING:0] str)
#  $swrite(uvm_bits_to_string, "%0s", str)
#endfunction


#verilator = True
verilator = False

sim_product = ''
if simulator.is_running():
    sim_product = simulator.get_simulator_product()
    print("uvm-python: Used simulator is |" + sim_product + "|")
    if sim_product == "Verilator":
        verilator = True

def uvm_has_verilator():
    return verilator

UVM_POUND_ZERO_COUNT = 1000
UVM_NO_WAIT_FOR_NBA = False

if hasattr(cocotb, 'SIM_NAME') and getattr(cocotb, 'SIM_NAME') == 'Verilator':
    UVM_POUND_ZERO_COUNT = 100


rw_event = ReadWrite()

async def uvm_wait_for_nba_region():
    """
    Task: uvm_wait_for_nba_region

    Callers of this task will not return until the `ReadWrite` region,
    thus allowing
    other processes any number of `NullTrigger`s to settle out before
    continuing. See `UVMSequencerBase.wait_for_sequences` for example usage.
    """
    if verilator is True:
        await rw_event
    else:
        if UVM_NO_WAIT_FOR_NBA is False:
            await rw_event
        else:
            for _ in range(0, UVM_POUND_ZERO_COUNT):
                await NullTrigger()


# Returns UVM coreservice
def get_cs():
    from .uvm_coreservice import UVMCoreService
    return UVMCoreService.get()

#// Class: uvm_enum_wrapper#(T)
#//
#// The ~uvm_enum_wrapper#(T)~ class is a utility mechanism provided
#// as a convenience to the end user.  It provides a <from_name>
#// method which is the logical inverse of the System Verilog ~name~
#// method which is built into all enumerations.

#class uvm_enum_wrapper#(type T=uvm_active_passive_enum)
#
#    protected static T map[string]
#
#    // Function: from_name
#    // Attempts to convert a string ~name~ to an enumerated value.
#    //
#    // If the conversion is successful, the method will return
#    // 1, otherwise 0.
#    //
#    // Note that the ~name~ passed in to the method must exactly
#    // match the value which would be produced by ~enum::name~, and
#    // is case sensitive.
#    //
#    // For example:
#    //| typedef uvm_enum_wrapper#(uvm_radix_enum) radix_wrapper
#    //| uvm_radix_enum r_v
#    //|
#    //| // The following would return '0', as "foo" isn't a value
#    //| // in uvm_radix_enum:
#    //| radix_wrapper::from_name("foo", r_v)
#    //|
#    //| // The following would return '0', as "uvm_bin" isn't a value
#    //| // in uvm_radix_enum (although the upper case "UVM_BIN" is):
#    //| radix_wrapper::from_name("uvm_bin", r_v)
#    //|
#    //| // The following would return '1', and r_v would be set to
#    //| // the value of UVM_BIN
#    //| radix_wrapper::from_name("UVM_BIN", r_v)
#    //
#    static function bit from_name(string name, ref T value)
#        if (map.size() == 0)
#          m_init_map()
#
#        if (map.exists(name)) begin
#            value = map[name]
#            return 1
#        end
#        else begin
#            return 0
#        end
#    endfunction : from_name
#
#    // Function- m_init_map
#    // Initializes the name map, only needs to be performed once
#    protected static function void m_init_map()
#        T e = e.first()
#        do
#          begin
#            map[e.name()] = e
#            e = e.next()
#          end
#        while (e != e.first())
#    endfunction : m_init_map
#
#    // Function- new
#    // Prevents accidental instantiations
#    def __init__(self, arr):
#       self.arr = arr
#
#
#endclass : uvm_enum_wrapper


"""
Group: uvm-python specific functions
"""

def uvm_is_sim_active():
    """
    Returns true if a simulator is active/attached. Returns False for example
    when running unit tests without cocotb Makefiles.
    """
    return simulator.is_running()


def uvm_sim_time(units='NS'):
    """
    Returns current simtime in the given units (default: NS)

    Args:
        units (str): PS, NS, US, MS or S
    Returns:
        int: Simulation time in specified units
    """
    if uvm_is_sim_active():
        return get_sim_time(units=units)
    return 0


async def uvm_zero_delay():
    await NullTrigger()


def uvm_check_output_args(arr):
    """
    Check that all args in the arr are lists to emulate the SV inout/output/ref args

    Raises:
    """
    for item in arr:
        if not isinstance(item, list):
            raise Exception('All output args must be given as empty arrays. Got: '
                    + str(item))
        elif len(item) != 0:
            raise Exception('All output args must be given as empty arrays. Got: '
                    + str(item) + ' len: ' + str(len(item)))



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#class producer #(type T=packet) extends uvm_component

import cocotb
from cocotb.triggers import Timer

from uvm.base.sv import sv
from uvm.base import (
        UVMComponent, UVMConfigDb, UVM_MEDIUM, UVM_NONE, UVM_HIGH, UVM_INFO)
from uvm.tlm1 import UVMBlockingPutPort
from uvm.macros import uvm_component_utils, uvm_info
from packet import packet


class producer(UVMComponent):
    #
    #  uvm_blocking_put_port #(T) out
    #
    def __init__(self, name, parent=None,T=packet):
        UVMComponent.__init__(self, name, parent)
        self.out = UVMBlockingPutPort("out", self)
        self.num_packets = 10
        nn = []
        self.T = T
        self.add_objection = True
        self.proto = self.T("my_packet")
        if UVMConfigDb.get(self, "", "num_packets", nn):
            self.num_packets = nn[0]

    
    async def run_phase(self, phase):
        if self.add_objection:
            phase.raise_objection(self)
        num = ""
        uvm_info("producer", "Starting with packet " + str(self.num_packets), UVM_MEDIUM)

        for count in range(self.num_packets):

            p = self.proto.clone()

            num = str(count)
            p.set_name(self.get_name() + "-" + num)
            p.set_initiator(self)

            if self.recording_detail != UVM_NONE:
                p.enable_recording(self.get_tr_stream("packet_stream"))

            p.randomize()
            uvm_info("producer", sv.sformatf("Sending %s",p.get_name()), UVM_MEDIUM)

            if self.uvm_report_enabled(UVM_HIGH,UVM_INFO,""):
                p.print_obj()

            await self.out.put(p)
            await Timer(10, "NS")

        uvm_info("producer", "Exiting.", UVM_MEDIUM)
        if self.add_objection:
            phase.drop_objection(self)


uvm_component_utils(producer)
#    `uvm_field_object(proto, UVM_ALL_ON + UVM_REFERENCE)
#    `uvm_field_int(num_packets, UVM_ALL_ON + UVM_DEC)
#    `uvm_field_int(count, UVM_ALL_ON + UVM_DEC + UVM_READONLY)
#  `uvm_component_utils_end



#
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

from typing import Dict

from .uvm_phase import UVMPhase
from .uvm_object_globals import UVM_PHASE_DOMAIN, UVM_PHASE_SCHEDULE
from .uvm_debug import uvm_debug
from ..macros import uvm_error


from .uvm_runtime_phases import (UVMMainPhase, UVMPreResetPhase,
    UVMResetPhase, UVMPostResetPhase, UVMPreConfigurePhase, UVMConfigurePhase,
    UVMPostConfigurePhase,
    UVMPreMainPhase, UVMPostMainPhase, UVMPreShutdownPhase, UVMShutdownPhase,
    UVMPostShutdownPhase)


# UVMPhases
build_ph = None
connect_ph = None
end_of_elaboration_ph = None
start_of_simulation_ph = None
run_ph = None
extract_ph = None
check_ph = None
report_ph = None



class UVMDomain(UVMPhase):
    """
    Class: UVMDomain

    Phasing schedule node representing an independent branch of the schedule.
    Handle used to assign domains to components or hierarchies in the testbench.
    Domain is basically a container for the uvm phases, or custom phases if
    custom schedule is created.
    """

    m_common_domain = None
    m_uvm_domain = None
    m_domains: Dict[str, 'UVMDomain'] = {}
    m_uvm_schedule = None  # type: UVMPhase

    @classmethod
    def get_domains(cls, domains=None) -> Dict[str, 'UVMDomain']:
        """
        Returns a list of all domains.

        Args:
            domains:
        Returns:
            list[UVMDomain]: List of all domains
        Raises:
            Exception
        """
        if domains is not None:
            raise Exception("get_domains() you must capture the retval")
        return cls.m_domains

    @classmethod
    def get_uvm_schedule(cls):
        """
        Get the "UVM" schedule, which consists of the run-time phases that
        all components execute when participating in the "UVM" domain.

        Returns:
            UVMPhase: Returns UVM schedule consisting of uvm phases.
        """
        cls.get_uvm_domain()
        return cls.m_uvm_schedule

    @classmethod
    def get_common_domain(cls):
        """
        Function: get_common_domain

        Get the "common" domain, which consists of the common phases that
        all components execute in sync with each other. Phases in the "common"
        domain are build, connect, end_of_elaboration, start_of_simulation, run,
        extract, check, report, and final.
        Returns:
        """
        domain = None
        schedule = None
        if UVMDomain.m_common_domain is not None:
            return UVMDomain.m_common_domain

        domain = UVMDomain("common")
        uvm_debug(cls, 'get_common_domain', 'Adding BuildPhase now')
        from .uvm_common_phases import (UVMBuildPhase, UVMConnectPhase,
                UVMEndOfElaborationPhase, UVMStartofSimulationPhase, UVMRunPhase,
                UVMExtractPhase, UVMCheckPhase, UVMReportPhase, UVMFinalPhase)
        domain.add(UVMBuildPhase.get())
        domain.add(UVMConnectPhase.get())
        domain.add(UVMEndOfElaborationPhase.get())
        domain.add(UVMStartofSimulationPhase.get())
        domain.add(UVMRunPhase.get())
        domain.add(UVMExtractPhase.get())
        domain.add(UVMCheckPhase.get())
        domain.add(UVMReportPhase.get())
        domain.add(UVMFinalPhase.get())
        UVMDomain.m_domains["common"] = domain

        # for backward compatibility, make common phases visible;
        # same as uvm_<name>_phase::get().
        uvm_debug(cls, 'get_common_domain', 'Before find() BuildPhase')
        build = UVMBuildPhase.get()
        build_ph               = domain.find(build)
        check_phase_exists('build', build_ph)
        connect_ph             = domain.find(UVMConnectPhase.get())
        check_phase_exists('connect', connect_ph)
        end_of_elaboration_ph  = domain.find(UVMEndOfElaborationPhase.get())
        check_phase_exists('end_of_elaboration', end_of_elaboration_ph)
        start_of_simulation_ph = domain.find(UVMStartofSimulationPhase.get())
        check_phase_exists('start_of_simulation', start_of_simulation_ph)
        run_ph                 = domain.find(UVMRunPhase.get())
        check_phase_exists('run', run_ph)
        extract_ph             = domain.find(UVMExtractPhase.get())
        check_phase_exists('extract', extract_ph)
        check_ph               = domain.find(UVMCheckPhase.get())
        check_phase_exists('check', check_ph)
        report_ph              = domain.find(UVMReportPhase.get())
        check_phase_exists('report', report_ph)
        UVMDomain.m_common_domain = domain

        domain = UVMDomain.get_uvm_domain()
        UVMDomain.m_common_domain.add(domain,
                with_phase=UVMDomain.m_common_domain.find(UVMRunPhase.get()))

        return UVMDomain.m_common_domain

    @classmethod
    def add_uvm_phases(cls, schedule):
        """
        Appends to the given `schedule` the built-in UVM phases.

        Args:
            schedule (UVMPhase): Phase to be added to the schedule.
        """
        schedule.add(UVMPreResetPhase.get())
        schedule.add(UVMResetPhase.get())
        schedule.add(UVMPostResetPhase.get())
        schedule.add(UVMPreConfigurePhase.get())
        schedule.add(UVMConfigurePhase.get())
        schedule.add(UVMPostConfigurePhase.get())
        schedule.add(UVMPreMainPhase.get())
        schedule.add(UVMMainPhase.get())
        schedule.add(UVMPostMainPhase.get())
        schedule.add(UVMPreShutdownPhase.get())
        schedule.add(UVMShutdownPhase.get())
        schedule.add(UVMPostShutdownPhase.get())

    @classmethod
    def get_uvm_domain(cls):
        """
        Get a handle to the singleton `uvm` domain

        Returns:
            UVMDomain: Singleton instance of uvm domain
        """
        if UVMDomain.m_uvm_domain is None:
            UVMDomain.m_uvm_domain = UVMDomain("uvm")
            UVMDomain.m_uvm_schedule = UVMPhase("uvm_sched", UVM_PHASE_SCHEDULE)
            UVMDomain.add_uvm_phases(UVMDomain.m_uvm_schedule)
            UVMDomain.m_uvm_domain.add(UVMDomain.m_uvm_schedule)
        return UVMDomain.m_uvm_domain

    def __init__(self, name):
        """
        Create a new instance of a phase domain.

        Args:
            name:
        """
        super().__init__(name, UVM_PHASE_DOMAIN)
        if name in UVMDomain.m_domains:
            uvm_error("UNIQDOMNAM",
                    "Domain created with non-unique name '{}'".format(name))
        UVMDomain.m_domains[name] = self

    # Function: jump
    #
    # jumps all active phases of this domain to to-phase if
    # there is a path between active-phase and to-phase
    #def jump(self, phase):
    #  uvm_phase phases[$];
    #
    #  m_get_transitive_children(phases);
    #
    #  phases = phases.find(item) with (item.get_state() inside {[UVM_PHASE_STARTED:UVM_PHASE_CLEANUP]});
    #
    #  foreach(phases[idx])
    #      if(phases[idx].is_before(phase) || phases[idx].is_after(phase))
    #          phases[idx].jump(phase);


    @classmethod
    def jump_all(cls, phase):
        """
        jump_all
        --------
        Args:
            phase:
        """
        #  uvm_domain domains[string];
        domains = UVMDomain.get_domains()  # string -> uvm_domain
        for idx in domains:
            domains[idx].jump(phase)

    #endclass UVMDomain

# tpoikela: Added for additional verification
def check_phase_exists(name, phase):
    if phase is None:
        msg = 'Phase ' + name + ' is None! in UVMDomain.get_common_domain()'
        raise Exception(msg)




from .uvm_agent import *
from .uvm_algorithmic_comparator import *
from .uvm_comps import *
from .uvm_driver import *
from .uvm_env import *
from .uvm_in_order_comparator import *
from .uvm_monitor import *
from .uvm_pair import *
from .uvm_policies import *
from .uvm_push_driver import *
from .uvm_random_stimulus import *
from .uvm_scoreboard import *
from .uvm_subscriber import *
from .uvm_test import *
